import logging
from collections import defaultdict
from typing import Dict, List, Tuple, Optional
from mip import Model, minimize, BINARY, xsum, OptimizationStatus, Var

from autobridge.Floorplan.Utilities import *
from autobridge.Opt.DataflowGraph import Vertex
from autobridge.Opt.Slot import Slot
from autobridge.Opt.SlotManager import SlotManager, Dir

_logger = logging.getLogger().getChild(__name__)


def _create_ilp_vars(
    m: Model, v_list: List[Vertex], s_list: List[Slot]
) -> Tuple[Dict[Vertex, Dict[Slot, Var]], Dict[Slot, Dict[Vertex, Var]]]:
  """
  for each vertex, for each slot, create a binary var if the vertex is assigned to the slot
  """
  v_to_s_to_var = defaultdict(dict)
  s_to_v_to_var = defaultdict(dict)
  for v in v_list:
    for s in s_list:
      v_to_s_to_var[v][s] = m.add_var(var_type=BINARY)
      s_to_v_to_var[s][v] = v_to_s_to_var[v][s]

  return v_to_s_to_var, s_to_v_to_var


def _add_area_constraints(
    m: Model, 
    s_to_v_to_var: Dict[Slot, Dict[Vertex, Var]],
    resource_usage_limit: int
) -> None:
  """
  limit the capacity of each slot
  """
  for r in RESOURCE_TYPES:
    for s, v_to_var in s_to_v_to_var.items():
      capacity = s.area[r] * resource_usage_limit
      m += xsum(v.area[r] * var for v, var in v_to_var.items()) <= capacity


def _add_unique_assign_constraints(m: Model, v_to_s_to_var: Dict[Vertex, Dict[Slot, Var]]) -> None:
  """
  each vertex is assigned to one slot
  """
  for v, s_to_var in v_to_s_to_var.items():
    m += xsum(var for var in s_to_var.values()) == 1


def _add_pre_assignments(
    m: Model, 
    v_to_s_to_var: Dict[Vertex, Dict[Slot, Var]],
    pre_assignments: Dict[Vertex, Slot],
    all_slot_list: List[Slot],
) -> None:
  for v, target_slot in pre_assignments.items():
    for candidate in all_slot_list:
      if not candidate.containsChildSlot(target_slot):
        m += v_to_s_to_var[v][candidate] == 0


def _get_v_to_s_to_cost(
    v_list: List[Vertex], s_list: List[Slot], orig_v2s: Dict[Vertex, Slot]
) -> Dict[Vertex, Dict[Slot, int]]:
  """
  cost for assigning a vertex to a slot
  Define the cost as the distance from the original location * the total wire length
  """
  v_to_s_to_cost = defaultdict(dict)
  for v in v_list:
    for s in s_list:
      orig_s = orig_v2s[v]
      # +1 to handle the case where a module has 0 wire length
      v_to_s_to_cost[v][s] = orig_s.getDistance(s) * (v.getTotalWireWidth() + 1)

  return v_to_s_to_cost


def _add_opt_goal(
    m: Model, 
    v_to_s_to_cost: Dict[Vertex, Dict[Slot, int]], 
    v_to_s_to_var: Dict[Vertex, Dict[Slot, Var]]
) -> None:
  """
  minimize the cost
  """
  cost_var_pair_list: List[Tuple[int, Var]] = []
  for v, s_to_var in v_to_s_to_var.items():
    for s, var in s_to_var.items():
      cost = v_to_s_to_cost[v][s]
      cost_var_pair_list.append((cost, var))

  m.objective = minimize(xsum(cost * var for cost, var in cost_var_pair_list))


def _add_grouping_constraints(
    m: Model,
    grouping_list: List[List[Vertex]],
    v_to_s_to_var: Dict[Vertex, Dict[Slot, Var]],
    s_list: List[Slot]
) -> None:

  for grouping in grouping_list:
    for i in range(1, len(grouping)):
      v1 = grouping[0]
      v2 = grouping[i]
      _logger.debug(f'Grouping {v1.name} and {v2.name}')
      for s in s_list:
        m += v_to_s_to_var[v1][s] == v_to_s_to_var[v2][s]


def _get_ilp_results(
    v_to_s_to_var: Dict[Vertex, Dict[Slot, int]]
) -> Tuple[ Dict[Vertex, Slot], Dict[Slot, List[Vertex]] ]:
  """
  extract which modules is assigned to which slots
  """
  # get v2s
  new_v2s = {}
  for v, s_to_var in v_to_s_to_var.items():
    for s, var in s_to_var.items():
      if round(var.x) == 1:
        new_v2s[v] = s
        break

    assert v in new_v2s

  # get s2v
  new_s2v = defaultdict(list)
  for v, s in new_v2s.items():
    new_s2v[s].append(v)

  return new_v2s, new_s2v


def _log_results(
    new_v2s: Dict[Vertex, Slot], 
    orig_v2s: Dict[Vertex, Slot]
) -> None:
  """
  analyze and log the new floorplan results
  """
  # log which vertices are re-placed
  for v, s in new_v2s.items():
    orig_s = orig_v2s[v]
    if s != orig_s:
      _logger.debug(f'Vertex {v.name} is moved {orig_s.getDistance(s)} units from {orig_s.getRTLModuleName()} to {s.getRTLModuleName()}')


def get_legalized_v2s(
    orig_v2s: Dict[Vertex, Slot],
    grouping_list: List[ List[Vertex] ],
    all_slot_list: List[Slot],
    pre_assignments: Dict[Vertex, Slot],
    resource_usage_limit: int
) -> Dict[Vertex, Slot]:
  """
  adjust the floorplanning to satisfy the area requirement
  """
  _logger.debug(f'Begin legalizing the floorplan results, target resource usage limit: {resource_usage_limit}')

  m = Model()
  m.verbose = 0

  v_list = list(orig_v2s.keys())
  s_list = all_slot_list

  v_to_s_to_var, s_to_v_to_var = _create_ilp_vars(m, v_list, s_list)
  v_to_s_to_cost = _get_v_to_s_to_cost(v_list, s_list, orig_v2s)

  _add_area_constraints(m, s_to_v_to_var, resource_usage_limit)

  _add_pre_assignments(m, v_to_s_to_var, pre_assignments, all_slot_list)

  _add_unique_assign_constraints(m, v_to_s_to_var)

  _add_grouping_constraints(m, grouping_list, v_to_s_to_var, s_list)

  _add_opt_goal(m, v_to_s_to_cost, v_to_s_to_var)

  status = m.optimize()
  if status != OptimizationStatus.OPTIMAL:
    _logger.debug(f'Fail to legalize the floorplan under target ratio {resource_usage_limit}')
    return {}

  new_v2s, new_s2v = _get_ilp_results(v_to_s_to_var)

  _log_results(new_v2s, orig_v2s)

  _logger.info('Finish legalizing the floorplan results.')

  return new_v2s


def legalize_floorplan(
  curr_v2s: Dict[Vertex, Slot],
  slot_manager: SlotManager,
  grouping_constraints: List[List[Vertex]],
  pre_assignments: Dict[Vertex, Slot],
  partition_order: List[Dir],
  init_usage_ratio: float = 0.7,
  limit_increase_step: float = 0.01,
  max_usage_ratio: float = 0.85,
) -> Optional[Dict[Vertex, Slot]]:
  """
  the iterative partitioning process may result in some slots being overused
  Re-assign some of the Vertices from the overused slots to the under-used slots
  """
  # if a slot is empty, it is not included in v2s
  all_leaf_slots = slot_manager.getLeafSlotsAfterPartition(partition_order)

  curr_limit = init_usage_ratio
  while 1:
    if curr_limit > max_usage_ratio:
      _logger.info(f'Fail to legalize under the cut threhold {max_usage_ratio}')
      return {}

    new_v2s = get_legalized_v2s(curr_v2s, grouping_constraints, all_leaf_slots, pre_assignments, curr_limit)
    if new_v2s:
      _logger.info(f'Legalization succeeded with target usage limit {curr_limit}')
      return new_v2s
    else:
      curr_limit += limit_increase_step


#! /usr/bin/python3.6
import sys
from collections import defaultdict
from autobridge.HLSParser.vivado_hls.HLSProjectManager import HLSProjectManager
from autobridge.HLSParser.vivado_hls.TopRTLParser import TopRTLParser
from autobridge.Device.DeviceManager import DeviceManager
from autobridge.Opt.DataflowGraph import DataflowGraph
from autobridge.Opt.Floorplan import Floorplanner
from autobridge.Opt.GlobalRouting import GlobalRouting
from autobridge.Opt.SlotManager import SlotManager
from autobridge.Opt.LatencyBalancing import LatencyBalancing
from autobridge.Codegen.Codegen import generate_new_rtl_top
from autobridge.Codegen.FloorplanTclGeneration import generate_floorplan_constraints
import logging
import json
import re
import os

class Manager:

  def __init__(
      self,
      config_file_path):
    assert os.path.isfile(config_file_path)
    self.config = json.loads(open(config_file_path, 'r').read())
    self.basicSetup()
    self.loggingSetup()

    hls_prj_manager = HLSProjectManager(self.top_rtl_name, self.hls_prj_path, self.hls_solution_name)
    top_rtl_parser = TopRTLParser(hls_prj_manager.getTopRTLPath())
    graph = DataflowGraph(hls_prj_manager, top_rtl_parser)

    slot_manager = SlotManager(self.board)

    user_constraint_s2v = self.parseUserConstraints(graph, slot_manager)

    floorplan = self.runFloorplanning(graph, user_constraint_s2v, slot_manager, hls_prj_manager, top_rtl_parser)

    # grid routing of edges 
    global_router = GlobalRouting(floorplan, top_rtl_parser, slot_manager)

    # latency balancing
    rebalance = LatencyBalancing(graph, floorplan, global_router)

    generate_new_rtl_top(top_rtl_parser.top_module_ast, graph.getNameToEdgeMap(), top_rtl_parser.getTopModuleName(), floorplan)
    generate_floorplan_constraints(floorplan, global_router)
  
  def basicSetup(self):
    if "BundleToDDRMapping" in self.config:
      # reserve area for ddrs
      ddr_list = list(set(self.config["BundleToDDRMapping"].values()))
      assert len(ddr_list) == len(self.config["BundleToDDRMapping"].values())
      self.device_manager = DeviceManager(self.config["Board"], ddr_list, True)
    else:
      self.device_manager = DeviceManager(self.config["Board"], [], True)
    
    self.board = self.device_manager.getBoard()

    self.top_rtl_name = self.config["TopName"]
    self.hls_prj_path = os.path.abspath(self.config["HLSProjectPath"])
    self.hls_solution_name = self.config["HLSSolutionName"]
    if "LoggingLevel" in self.config:
      self.logging_level = self.config["LoggingLevel"]
    else:
      self.logging_level = "INFO"

    if "Target" in self.config:
      self.target = self.config["Target"]
    else:
      self.target = "hw"

  def loggingSetup(self):
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    formatter = logging.Formatter("[%(levelname)s: %(funcName)25s() ] %(message)s")
    
    debug_file_handler = logging.FileHandler(filename='autobridge-debug.log', mode='w')
    debug_file_handler.setLevel(logging.DEBUG)
    info_file_handler = logging.FileHandler(filename='autobridge-info.log', mode='w')
    info_file_handler.setLevel(logging.INFO)
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(logging.INFO)

    handlers = [debug_file_handler, info_file_handler, stdout_handler]
    for handler in handlers:
      handler.setFormatter(formatter)
      root.addHandler(handler)

  def parseUserConstraints(self, graph, slot_manager):
    user_constraint_s2v = defaultdict(list)
    
    def get_interface_module_name(bundle_name):
      top_name = self.config["TopName"]
      return f'{top_name}_{bundle_name}_m_axi_U'

    def get_s_axi_control_name():
      top_name = self.config["TopName"]
      return f'{top_name}_control_s_axi_U'

    # constrain the s_axi_control to SLR 0
    CONTROL_S_AXI_LOCATION = 0

    if "BundleToDDRMapping" in self.config:
      # properly constrain the interface modules
      board = self.device_manager.getBoard()
      for bundle, ddr in self.config["BundleToDDRMapping"].items():
        ddr_pblock = board.getDDRPblock(ddr)
        slot = slot_manager.createSlot(ddr_pblock)
        v_name = get_interface_module_name(bundle)
        try:
          v = graph.getVertex(v_name)
        except:
          logging.error(f'Naming convention of interface modules has changed!')
          exit(1)

        user_constraint_s2v[slot].append(v)
        logging.info(f'Constrain interface module {v_name} to {ddr_pblock}')

      # constrain the s_axi_control
      s_axi_slot = slot_manager.createSlot(board.getDDRPblock(CONTROL_S_AXI_LOCATION))
      try:
        s_axi_vertex = graph.getVertex(get_s_axi_control_name())
      except:
        logging.error(f'Naming convention of s_axi_control has changed!')
        exit(1)
      user_constraint_s2v[s_axi_slot].append(s_axi_vertex)
      logging.info(f'Constrain s_axi_control {v_name} to {board.getDDRPblock(CONTROL_S_AXI_LOCATION)}')

    if "Floorplan" in self.config:
      user_fp_json = self.config["Floorplan"]
      for region, v_name_group in user_fp_json.items():
        slot = slot_manager.createSlot(region)
        for v_name in v_name_group:
          user_constraint_s2v[slot].append(graph.getVertex(v_name))

    return user_constraint_s2v

  def runFloorplanning(
      self, 
      graph, 
      user_constraint_s2v, 
      slot_manager, 
      hls_prj_manager,
      top_rtl_parser,
      grouping_constraints=[]):

    grouping_constraints += top_rtl_parser.getStrictGroupingConstraints()
    
    floorplan = Floorplanner(
      graph, 
      user_constraint_s2v, 
      slot_manager=slot_manager, 
      total_usage=hls_prj_manager.getTotalArea(), 
      board=self.device_manager.getBoard(),
      user_max_usage_ratio=self.config['AreaUtilizationRatio'],
      grouping_constraints=grouping_constraints)
    
    if 'FloorplanMethod' in self.config:
      choice = self.config['FloorplanMethod']
      if choice == 'IterativeDivisionToFourCRs':
        floorplan.naiveFineGrainedFloorplan()
      elif choice == 'IterativeDivisionToHalfSLR':
        floorplan.coarseGrainedFloorplan()
      elif choice == 'IterativeDivisionToTwoCRs':
        floorplan.naiveTwoCRGranularityFloorplan()
      elif choice == 'EightWayDivisionToHalfSLR':
        floorplan.eightWayPartition()
      else:
        assert False, f'unsupported floorplan method: {choice}'
    else:
      floorplan.coarseGrainedFloorplan() # by default

    floorplan.printFloorplan()
    return floorplan

  def help(self):
    manual = {
      "Board" : "Choose between U250 and U280",
      "TopName" : "The name of the top-level function of the HLS design",
      "HLSProjectPath" : "Absolute path to the post-csynth HLS project",
      "HLSSolutionName" : "Name of the solution of the HLS project",
      "FloorplanMethod" : "Choose between IterativeDivisionToFourCRs and IterativeDivisionToHalfSLR",
      "AreaUtilizationRatio" : "Allowed resource usage in each slot",
      "Floorplan (optional)" : {
        "This is a comment" : "User could dictate the final location of given modules",
        "Region" : [
          "RTL module 1",
          "RTL module 2"
        ]
      },
      "LoggingLevel (optional)": "Choose between DEBUG, INFO, WARNING, CRITICAL, ERROR"
    }
    print(json.dumps(manual, indent=2))

if __name__ == "__main__":
  assert len(sys.argv) == 2, 'input the path to the configuration file'
  m = Manager(sys.argv[1])
    





from collections import defaultdict
from typing import List, Dict

from autobridge.Opt.Slot import Slot


def get_ap_start_pipeline_def(slot_list: List[Slot], pipeline_level=4) -> List[str]:
  pipeline_def = []
  pipeline_def.append(f'// ----- pipelining the ap_start signal -----')
  
  pipeline_def.append(f'wire ap_start_orig;')

  for slot in slot_list:
    pipeline_def.append(f'(* keep = "true" *) reg ap_start_{slot.getRTLModuleName()};')
    for level in range(pipeline_level):
      pipeline_def.append(f'(* keep = "true" *) reg ap_start_{slot.getRTLModuleName()}_q{level};')

    pipeline_def.append(f'always @ (posedge ap_clk) begin')
    pipeline_def.append(f'  ap_start_{slot.getRTLModuleName()}_q0 <= ap_start_orig;')
    for level in range(1, pipeline_level):
      pipeline_def.append(f'  ap_start_{slot.getRTLModuleName()}_q{level} <= ap_start_{slot.getRTLModuleName()}_q{level-1};')
    pipeline_def.append(f'  ap_start_{slot.getRTLModuleName()} <= ap_start_{slot.getRTLModuleName()}_q{pipeline_level-1};')
    pipeline_def.append(f'end')

    pipeline_def.append(f'\n')

  pipeline_def.append(f'// ----- end of pipelining the ap_start signal -----\n')

  # add indentation
  pipeline_def = ['  ' + stmt for stmt in pipeline_def]
  
  return pipeline_def


def get_ap_rst_pipeline_def(slot_list: List[Slot], pipeline_level=4) -> List[str]:
  pipeline_def = []
  pipeline_def.append(f'// ----- pipelining the ap_rst signal -----')

  for slot in slot_list:
    pipeline_def.append(f'(* keep = "true" *) reg ap_rst_{slot.getRTLModuleName()};')
    for level in range(pipeline_level):
      pipeline_def.append(f'(* keep = "true" *) reg ap_rst_{slot.getRTLModuleName()}_q{level};')

    pipeline_def.append(f'always @ (posedge ap_clk) begin')
    pipeline_def.append(f'  ap_rst_{slot.getRTLModuleName()}_q0 <= !ap_rst_n;')
    for level in range(1, pipeline_level):
      pipeline_def.append(f'  ap_rst_{slot.getRTLModuleName()}_q{level} <= ap_rst_{slot.getRTLModuleName()}_q{level-1};')
    pipeline_def.append(f'  ap_rst_{slot.getRTLModuleName()} <= ap_rst_{slot.getRTLModuleName()}_q{pipeline_level-1};')
    pipeline_def.append(f'end')

    pipeline_def.append(f'\n')

  pipeline_def.append(f'// ----- end of pipelining the ap_rst_ signal -----\n')

  # add indentation
  pipeline_def = ['  ' + stmt for stmt in pipeline_def]
  
  return pipeline_def


def get_ap_done_pipeline_def(v_name_to_s: Dict[str, Slot], ap_done_module_list: List[str]) -> List[str]:
  pipeline_def = []
  pipeline_def.append(f'// ----- pipelining the ap_done signal -----')

  # get the union ap_done from each slot
  # note that some modules dont have ap_done
  s_to_v_name_list = defaultdict(list)
  for v_name in ap_done_module_list:
    slot = v_name_to_s[v_name]
    s_to_v_name_list[slot].append(v_name)

  # reigster the ap_dones from each module instance
  # hold the signal unless reset
  for v_name in ap_done_module_list:
    slot = v_name_to_s[v_name]
    pipeline_def.append(f'wire ap_done_{v_name};')
    pipeline_def.append(f'(* keep = "true" *) reg ap_done_{v_name}_q0;')
  pipeline_def.append(f'\n')

  for slot, v_name_list in s_to_v_name_list.items():
    pipeline_def.append(f'(* keep = "true" *) reg ap_done_{slot.getRTLModuleName()}_q0;')
    pipeline_def.append(f'(* keep = "true" *) reg ap_done_{slot.getRTLModuleName()}_q1;')
    pipeline_def.append(f'(* keep = "true" *) reg ap_done_{slot.getRTLModuleName()}_q2;')
  pipeline_def.append(f'\n')

  pipeline_def.append(f'(* keep = "true" *) reg ap_done_final;')
  for slot, v_name_list in s_to_v_name_list.items():
    pipeline_def.append(f'(* keep = "true" *) reg ap_done_final_{slot.getRTLModuleName()}_q0;')
    pipeline_def.append(f'(* keep = "true" *) reg ap_done_final_{slot.getRTLModuleName()}_q1;')
    pipeline_def.append(f'(* keep = "true" *) reg ap_done_final_{slot.getRTLModuleName()}_q2;')
  pipeline_def.append(f'\n')

  # get the ap_done of a slot. Reset the ap_done of each module once the slot ap_done is captured
  for v_name in ap_done_module_list:
    slot = v_name_to_s[v_name]
    pipeline_def.append(f'always @ (posedge ap_clk) ap_done_{v_name}_q0 <= (ap_done_{v_name} | ap_done_{v_name}_q0) & (!ap_rst_{slot.getRTLModuleName()}) & (!ap_done_{slot.getRTLModuleName()}_q1);')
  pipeline_def.append(f'\n')

  # get the final ap_done. Reset the ap_done of each slot once the final ap_done is captured
  for slot, v_name_list in s_to_v_name_list.items():
    pipeline_def.append(f'always @ (posedge ap_clk) ap_done_{slot.getRTLModuleName()}_q0 <= ' + ' & '.join([f'ap_done_{v_name}_q0' for v_name in v_name_list]) + ';')
    pipeline_def.append(f'always @ (posedge ap_clk) ap_done_{slot.getRTLModuleName()}_q1 <= ap_done_{slot.getRTLModuleName()}_q0;')
    pipeline_def.append(f'always @ (posedge ap_clk) ap_done_{slot.getRTLModuleName()}_q2 <= (ap_done_{slot.getRTLModuleName()}_q2 | ap_done_{slot.getRTLModuleName()}_q1) & (!ap_rst_{slot.getRTLModuleName()}) & (!ap_done_final_{slot.getRTLModuleName()}_q2);')
    pipeline_def.append(f'\n')

  # get the final ap_done
  pipeline_def.append(f'assign ap_done = ap_done_final;')  # for compatibility with HLS simulation
  pipeline_def.append(f'always @ (posedge ap_clk) ap_done_final <= ' + ' & '.join([f'ap_done_{slot.getRTLModuleName()}_q2' for slot in s_to_v_name_list.keys()]) + ';')
  pipeline_def.append(f'\n')

  # pipeline the final ap_done back to each slot to reset the slot-level ap_done
  for slot, v_name_list in s_to_v_name_list.items():
    pipeline_def.append(f'always @ (posedge ap_clk) ap_done_final_{slot.getRTLModuleName()}_q0 <= ap_done_final;')
    pipeline_def.append(f'always @ (posedge ap_clk) ap_done_final_{slot.getRTLModuleName()}_q1 <= ap_done_final_{slot.getRTLModuleName()}_q0;')
    pipeline_def.append(f'always @ (posedge ap_clk) ap_done_final_{slot.getRTLModuleName()}_q2 <= ap_done_final_{slot.getRTLModuleName()}_q1;')

  pipeline_def.append(f'// ----- end of pipelining the ap_done signal -----\n')

  # add indentation
  pipeline_def = ['  ' + stmt for stmt in pipeline_def]
  
  return pipeline_def


def remove_orig_ctrl_signal(temp_rtl_top: List[str]) -> None:
  """
  comment out all assign statement of the original control signals
  also comment out the always blocks related to the original control signals
  """
  ctrl_signals = [
    'ap_start',
    'ap_ready',
    'ap_done',
    'ap_continue',
    'ap_idle',
    'ap_sync_',
  ]

  # the always blocks are suppose to be after all module instances
  for i in range(len(temp_rtl_top)):
    if 'always' in temp_rtl_top[i] and 'ap_clk' in temp_rtl_top[i]:
      for j in range(i, len(temp_rtl_top)):
        if 'assign' in temp_rtl_top[j]:
          break
        else:
          temp_rtl_top[j] = f'//  {temp_rtl_top[j]}'
      break

  # all assign statement are after the always blocks
  for i in range(len(temp_rtl_top)):
    if 'assign' in temp_rtl_top[i]:
      assert any(ctrl_signal in temp_rtl_top[i] for ctrl_signal in ctrl_signals), temp_rtl_top[i]
      temp_rtl_top[i] = f'//  {temp_rtl_top[i]}'



import logging

from typing import Dict, List, Optional
from mip import Model, Var, minimize, xsum, BINARY, INTEGER, OptimizationStatus

from autobridge.Floorplan.Utilities import *
from autobridge.Opt.DataflowGraph import Vertex
from autobridge.Opt.Slot import Slot
from autobridge.Opt.SlotManager import SlotManager, Dir

_logger = logging.getLogger().getChild(__name__)


class Bipartition:
  def __init__(
    self,
    curr_v2s: Dict[Vertex, Slot],
    grouping_constraints: List[List[Vertex]],
    pre_assignments: Dict[Vertex, Slot],
    slot_manager: SlotManager,
  ):
    """
    grouping_constraints: which modules should be assigned to the same slot
    pre_assignment: pre-determined module -> slot mapping
    """
    self.grouping_constraints = grouping_constraints
    self.pre_assignment = pre_assignments
    self.slot_manager = slot_manager

    self.set_curr_v2s(curr_v2s)

  def set_curr_v2s(self, curr_v2s: Dict[Vertex, Slot]):
    self.curr_v2s = curr_v2s
    self.curr_s2v = invert_v2s(curr_v2s)

  def get_bipartition(
      self, 
      direction: Dir,
      max_usage_ratio: float,
      max_search_time: int,
  ) -> Optional[Dict[Vertex, Slot]]:
    """
    bi-partition all the current slots
    """
    m = Model()
    m.verbose = 0

    v2var = self._create_ilp_vars(m)

    self._add_opt_goal(m, v2var, direction)

    # area constraints for each child slot
    self._add_area_constraints(m, v2var, direction, max_usage_ratio)

    self._add_pre_assignment(m, v2var, direction)

    self._add_grouping_constraints(m, v2var)

    m.optimize(max_seconds=max_search_time)

    next_v2s = self._get_partition_result(m, v2var, direction)
    if not next_v2s:
      _logger.debug(f'bi-partioning failed with usage ratio {max_usage_ratio}')

    return next_v2s

  def get_bipartition_adjust_ratio(
      self,
      direction: Dir,
      ref_usage_ratio: float,
      usage_ratio_delta: float = 0.02,
      max_search_time: int = 120
  ) -> Dict[Vertex, Slot]:
    """
    increase the ref_usage_ratio until succeed. Each time increment usage_ratio_delta
    and retry.
    """
    curr_usage_ratio = ref_usage_ratio

    while 1:
      _logger.debug(f'Attempt bipartition with a usage ratio of {curr_usage_ratio} and max search time {max_search_time}s')
      next_v2s = self.get_bipartition(direction, curr_usage_ratio, max_search_time)
      if not next_v2s:
        curr_usage_ratio += usage_ratio_delta
        curr_usage_ratio = round(curr_usage_ratio, 2)
      else:
        break
    _logger.info(f'Finish bipartition with a usage ratio of {curr_usage_ratio}')

    return next_v2s

  ############################################################################

  def _create_ilp_vars(self, m: Model):
    """
    create a binary var for each vertex
    the var indicates if a vertex is assigned to the up/right child slot or the down/left child slot
    """
    v2var: Dict[str, Var] = {} # str -> [mip_var]
    for v in self.curr_v2s.keys():
      v2var[v] = m.add_var(var_type=BINARY, name=f'{v.name}_x')
    return v2var

  def _add_opt_goal(
      self, 
      m: Model, 
      v2var: Dict[str, Var], 
      direction: Dir
  ) -> None:
    """
    minimize the weighted sum over all edges
    """
    edge_list = get_all_edges(list(self.curr_v2s.keys()))
    e2cost_var = {
      e: m.add_var(var_type=INTEGER, name=f'e_cost_{e.name}') for e in edge_list
    }

    def _get_loc_after_partition(v: Vertex):
      if direction == Dir.vertical:
        return self.curr_v2s[v].getQuarterPositionX() + v2var[v] * self.curr_v2s[v].getHalfLenX()
      elif direction == Dir.horizontal:
        return self.curr_v2s[v].getQuarterPositionY() + v2var[v] * self.curr_v2s[v].getHalfLenY()
      else:
        assert False

    for e, cost_var in e2cost_var.items():
      m += cost_var >= _get_loc_after_partition(e.src) - _get_loc_after_partition(e.dst)
      m += cost_var >= _get_loc_after_partition(e.dst) - _get_loc_after_partition(e.src)

    m.objective = minimize( xsum(cost_var * e.width for e, cost_var in e2cost_var.items()) )

  def _add_area_constraints(
      self, 
      m: Model, 
      v2var: Dict[str, Var], 
      direction: Dir,
      max_usage_ratio: float,
  ) -> None:
    for s, v_group in self.curr_s2v.items():
      bottom_or_left, up_or_right = self.slot_manager.partitionSlotByHalf(s, direction)

      for r in RESOURCE_TYPES:
        v_var_list = [v2var[v] for v in v_group]
        area_list = [v.area[r] for v in v_group]
        I = range(len(v_group))

        # for the up/right child slot (if mod_x is assigned 1)
        m += xsum( v_var_list[i] * area_list[i] for i in I ) <= up_or_right.getArea()[r] * max_usage_ratio

        # for the down/left child slot (if mod_x is assigned 0)
        m += xsum( (1-v_var_list[i]) * area_list[i] for i in I ) <= bottom_or_left.getArea()[r] * max_usage_ratio

  def _add_pre_assignment(
      self,
      m: Model,
      v2var: Dict[str, Var], 
      direction: Dir,
  ) -> None:
    """
    user specifies that certain Vertices must be assigned to certain slots
    """
    for v, expect_slot in self.pre_assignment.items():
      curr_slot = self.curr_v2s[v]
      bottom_or_left, up_or_right = self.slot_manager.partitionSlotByHalf(curr_slot, direction)
      if bottom_or_left.containsChildSlot(expect_slot):
        _logger.debug(f'[specified assignment] {v.name} assigned to bottom/left')
        m += v2var[v] == 0
      elif up_or_right.containsChildSlot(expect_slot):
        _logger.debug(f'[specified assignment] {v.name} assigned to up/right')
        m += v2var[v] == 1
      else:
        _logger.warning(f'Potential wrong constraints from user: {v.name} -> {expect_slot.getName()}')
        assert False

  def _add_grouping_constraints(
      self, 
      m: Model,
      v2var: Dict[str, Var], 
  ) -> None:
    """
    user specifies that certain Vertices must be assigned to the same slot
    """
    for grouping in self.grouping_constraints:
      for i in range(1, len(grouping)):
        m += v2var[grouping[0]] == v2var[grouping[i]]

  def _get_partition_result(
      self,
      m: Model,
      v2var: Dict[str, Var], 
      direction: Dir,
  ) -> Optional[Dict[Vertex, Slot]]:
    """
    extract the results from the ILP solver
    """
    if m.status != OptimizationStatus.OPTIMAL and m.status != OptimizationStatus.FEASIBLE:
      return {}

    # create new mapping
    next_v2s = {}
    for s, v_group in self.curr_s2v.items():
      bottom_or_left, up_or_right = self.slot_manager.partitionSlotByHalf(s, direction)

      for v in v_group:
        # if v is assigned to 0-half in the i-th solution
        if round(v2var[v].x) == 0:
          next_v2s[v] = bottom_or_left
        # if v is assigned to 1-half in the i-th solution
        elif round(v2var[v].x) == 1:
          next_v2s[v] = up_or_right

        else:
          assert False, v2var[v].x

    return next_v2s



import logging


class GlobalRouting:
  def __init__(self, floorplan, top_rtl_parser, slot_manager):
    self.floorplan = floorplan
    self.top_rtl_parser = top_rtl_parser
    self.slot_manager = slot_manager
    self.v2s = floorplan.getVertexToSlot()
    self.s2e = floorplan.getSlotToEdges()
    self.e_name2path = {} # from edge to all slots passed

    self.naiveGlobalRouting()
    self.updateEdgePipelineLevel()

  def updateEdgePipelineLevel(self):
    """
    update the pipeline_level filed based on the routing results
    """
    for e_list in self.s2e.values():
      for e in e_list:
        slot_path = []
        src_slot = self.v2s[e.src]
        dst_slot = self.v2s[e.dst]
        slot_path = self.e_name2path[e.name]

        # 2 levels of pipelining for each slot crossing
        if src_slot == dst_slot:
          e.pipeline_level = 0
        else:
          e.pipeline_level = (len(slot_path) + 1) * 2

  def naiveGlobalRouting(self):
    """
    each edge first go in the Y direction then in the X direction
    assume all slots are of the same size and are aligned
    the slot_path exclude the src slot and the dst slot
    """
    for e_list in self.s2e.values():
      for e in e_list:
        slot_path = []
        src_slot = self.v2s[e.src]
        dst_slot = self.v2s[e.dst]
        slot_path.append(src_slot)

        curr = src_slot
        len_x = src_slot.getLenX()
        len_y = src_slot.getLenY()

        # first go in X direction
        x_diff = curr.getPositionX() - dst_slot.getPositionX()
        if x_diff:
          dir = 'LEFT' if x_diff > 0 else 'RIGHT'
          for i in range(int(abs(x_diff/len_x))):
            curr = self.slot_manager.createSlotForRouting(curr.getNeighborSlotName(dir))
            slot_path.append(curr)

        y_diff = curr.getPositionY() - dst_slot.getPositionY()
        if y_diff:
          dir = 'DOWN' if y_diff > 0 else 'UP'
          for i in range(int(abs(y_diff/len_y))):
            curr = self.slot_manager.createSlotForRouting(curr.getNeighborSlotName(dir))
            slot_path.append(curr)
        
        assert curr == dst_slot
        
        slot_path = slot_path[1:-1] # exclude the src and the dst
        logging.info(f'{e.name}: {self.v2s[e.src].getName()} -> {self.v2s[e.dst].getName()} : ' + ' '.join(s.getName() for s in slot_path))
        self.e_name2path[e.name] = slot_path



import logging
import math
from typing import Dict

from autobridge.HLSParser.vivado_hls.TopRTLParser import TopRTLParser
from autobridge.HLSParser.vivado_hls.HLSProjectManager import HLSProjectManager

FIFO_TYPE_DEPTH_THRESHOLD = 1024


class Edge:
  def __init__(self, name:str):
    self.src : Vertex = None
    self.dst : Vertex = None
    self.width = -1
    self.depth = -1  # original depth as in the RTL generated by HLS
    self.addr_width = -1
    self.name = name
    self.pipeline_level = 0
    self.added_depth_for_rebalance = 0
    self.fifo_type = ""

  def __hash__(self):
    return hash(self.name)

  def __eq__(self, other):
    return self.name == other.name

  def setDepth(self, depth):
    self.depth = depth
    self.addr_width = int(math.log2(self.depth)+1)

  def addDepth(self, depth_delta):
    self.setDepth(self.depth + depth_delta)

  def setWidth(self, width):
    self.width = width

  def getDefaultType(self) -> str:
    if self.depth >= FIFO_TYPE_DEPTH_THRESHOLD:
      return 'BRAM'
    else:
      return 'SRL'

  def getSRLFIFOArea(self, skip_balance_part: bool) -> Dict[str, int]:
    fifo_depth = self.depth
    if not skip_balance_part:
      fifo_depth += self.added_depth_for_rebalance

    lut_ram = self.width * ((fifo_depth - 1) / 16 + 1)
    lut_logic = 15 + 3 * int(math.log2(fifo_depth))
    ff = 7 + 3 * int(math.log2(fifo_depth))

    return {
      'BRAM': 0,
      'DSP': 0,
      'URAM': 0,
      'FF': ff,
      'LUT': lut_ram + lut_logic
    }

  def getBRAMFIFOArea(self, skip_balance_part: bool) -> Dict[str, int]:
    return {
      'BRAM': int(self.width/36 * 2),  # x2 because we are counting BRAM_18k
      'DSP': 0,
      'URAM': 0,
      'FF': self.width,  # approximation
      'LUT': 0  # based on HLS report
    }

  def getArea(self, skip_balance_part: bool = False) -> Dict[str, int]:
    if self.fifo_type == '':
      fifo_type = self.getDefaultType()
    else:
      fifo_type = self.fifo_type

    if fifo_type == 'BRAM':
      return self.getBRAMFIFOArea(skip_balance_part)
    elif fifo_type == 'SRL':
      return self.getSRLFIFOArea(skip_balance_part)
    else:
      assert False, fifo_type


class Vertex():
  def __init__(self, type:str, name : str):
    self.in_edges = [] # stores Edge objects
    self.out_edges = []
    self.in_edge_names = [] # stores Edge objects
    self.out_edge_names = []
    self.type = type
    self.name = name
    self.id = self.type + self.name
    self.area = {} # str_name -> count

  def __hash__(self):
    return hash(self.id)

  def __eq__(self, other):
    return self.id == other.id

  def getEdgeNames(self):
    return self.in_edge_names + self.out_edge_names

  def getEdges(self):
    return self.in_edges + self.out_edges
  
  def getInEdges(self):
    return self.in_edges
  
  def getOutEdges(self):
    return self.out_edges

  def getNeighborVertices(self):
    src_neighbors = [e.src for e in self.getInEdges()]
    dst_neighbors = [e.dst for e in self.getOutEdges()]
    return src_neighbors + dst_neighbors

  def getTotalWireWidth(self) -> int:
    """
    sum of the width of all edges of the vertex
    """
    return sum(e.width for e in self.getEdges())

  def getVertexAndInboundFIFOArea(self) -> Dict[str, int]:
    return {
      item: sum(e.getArea()[item] for e in self.in_edges) + val for item, val in self.area.items()
    }
  

class DataflowGraph:
  def __init__(self, hls_prj_manager : HLSProjectManager, top_rtl_parser : TopRTLParser):
    self.hls_prj_manager = hls_prj_manager
    self.top_rtl_parser = top_rtl_parser

    self.vertices = {} # name -> Vertex
    self.edges = {} # name -> Edge

    for v_node in self.top_rtl_parser.traverseVertexInAST():
      self.__initVertices(v_node)

    for e_node in self.top_rtl_parser.traverseEdgeInAST():
      self.__initEdges(e_node)

    self.__linkEdgeAndVertex()
    
    self.__checker()

    self.v_type_2_int = {} # map each vertex type to an integer
    self.v_name_2_int = {} # map each vertex instance to an integer
    self.int_2_v_type = {}
    self.int_2_v_name = {}
    self.__initVTypeToInt()
    self.__initVInstToInt()
    
  # assign an integer to v type
  def __initVTypeToInt(self):
    id = 1
    for v_node in self.vertices.values():
      if v_node.type not in self.v_type_2_int:
        self.v_type_2_int[v_node.type] = id
        self.int_2_v_type[id] = v_node.type
        logging.debug(f'{v_node.type} : {id}')
        id += 1

  # assign an integer to v name
  def __initVInstToInt(self):
    id = 1
    for v_node in self.vertices.values():
      if v_node.name not in self.v_name_2_int:
        self.v_name_2_int[v_node.name] = id
        self.int_2_v_name[id] = v_node.name
        logging.debug(f'{v_node.name} : {id}')
        id += 1
      else:
        assert False, f'Found two modules of the same name {v_node.name}'

  # note that we need to merge repetitve edges
  def getIntegerGraph(self):
    int_e_list = [ (self.v_name_2_int[e.src.name], self.v_name_2_int[e.dst.name]) \
                  for e in self.edges.values() ]
    int_e_list = list(set(int_e_list))

    int_v_labels = [ [self.v_name_2_int[v.name], self.v_type_2_int[v.type] ] \
                  for v in self.vertices.values()]
    return int_e_list, int_v_labels

  def getIntEdgeToName(self):
    int_edge2name = { (self.v_name_2_int[e.src.name], self.v_name_2_int[e.dst.name]) : e.name \
                  for e in self.edges.values() }
    return int_edge2name

  def getIntIdToVType(self):
    return self.int_2_v_type

  def getIntIdToVName(self):
    return self.int_2_v_name

  def __checker(self):
    v_name_list = [v.type + v.name for v in self.getAllVertices()]
    e_name_list = [e.name for e in self.getAllEdges()]
    assert len(v_name_list) == len(set(v_name_list)), 'Find repeated modules'
    assert len(e_name_list) == len(set(e_name_list))

  def __initVertices(self, v_node):

    v = Vertex(v_node.module, v_node.name)

    # get area
    v.area = self.hls_prj_manager.getAreaFromModuleType(v.type)
    
    v.in_edge_names = self.top_rtl_parser.getInFIFOsOfModuleInst(v.name)
    v.out_edge_names = self.top_rtl_parser.getOutFIFOsOfModuleInst(v.name)

    self.vertices[v_node.name] = v

  def __initEdges(self, e_node):

    e = Edge(e_node.name)

    # extract width
    e.setWidth(self.top_rtl_parser.getFIFOWidthFromFIFOType(e_node.module))
    e.setDepth(self.top_rtl_parser.getFIFODepthFromFIFOType(e_node.module))

    self.edges[e_node.name] = e

  def __linkEdgeAndVertex(self):
    for v in self.vertices.values():
      for fifo_in_name in v.in_edge_names:
        fifo_in = self.edges[fifo_in_name]
        fifo_in.dst = v
        v.in_edges.append(fifo_in)
      for fifo_out_name in v.out_edge_names:
        fifo_out = self.edges[fifo_out_name]
        fifo_out.src = v
        v.out_edges.append(fifo_out)

  def printVertices(self):
    for v in self.vertices.values():
      logging.debug(f'{v.name}: {v.area}')
      for e in v.in_edges:
        logging.debug(f'  <- {e.name}')
      for e in v.out_edges:
        logging.debug(f'  -> {e.name}')

  def printEdges(self):
    for e in self.edges.values():
      logging.debug(f'{e.name}: {e.src.name} -> {e.dst.name}')

  def getAllVertices(self):
    return self.vertices.values()

  def getAllEdges(self):
    return self.edges.values()

  def getNameToVertexMap(self):
    return self.vertices

  def getNameToEdgeMap(self):
    return self.edges

  def getVertex(self, v_name):
    return self.vertices[v_name]


import logging
import logging.handlers
import os

from datetime import datetime
from typing import Dict, List
from autobridge.Floorplan.EightWayPartition import eight_way_partition
from autobridge.Floorplan.FourWayPartition import four_way_partition
from autobridge.Floorplan.IterativeBipartion import iterative_bipartition
from autobridge.Floorplan.Utilities import print_pre_assignment, print_vertex_areas
from autobridge.Opt.DataflowGraph import Vertex, DataflowGraph
from autobridge.Opt.Slot import Slot
from autobridge.Opt.SlotManager import SlotManager, Dir

log_file = f'tapa-autobridge-{datetime.now().strftime("%d-%m-%Y-%H-%M")}.log'
handler = logging.FileHandler(log_file)

handlers = [handler, logging.StreamHandler()]
for h in handlers:
  h.setFormatter(logging.Formatter('%(message)s'))
logging.basicConfig(level = logging.INFO, handlers = handlers)
_logger = logging.getLogger().getChild(__name__)


def get_floorplan(
  graph: DataflowGraph,
  slot_manager: SlotManager,
  grouping_constraints_in_str: List[List[str]],
  pre_assignments_in_str: Dict[str, str],
  partition_order_in_str: List[str] = ['HORIZONTAL', 'HORIZONTAL', 'VERTICAL'],
  ref_usage_ratio: float = 0.7,
  threshold_for_iterative: int = 400,
) -> Dict[Vertex, Slot]:
  """
  main entrance of the floorplan part
  """
  # get initial v2s
  init_slot = slot_manager.getInitialSlot()
  init_v2s = {v : init_slot for v in graph.getAllVertices()}

  # get grouping constraints of Vertex
  grouping_constraints: List[List[Vertex]] = [
    [graph.getVertex(v_name) for v_name in v_name_group]
      for v_name_group in grouping_constraints_in_str
  ]

  _logger.info(f'The following modules are grouped to the same location:')
  for grouping in grouping_constraints_in_str:
    _logger.info('    ' + ', '.join(grouping))

  # get pre_assignment in Vertex
  pre_assignments = { graph.getVertex(v_name) : slot_manager.createSlot(pblock) 
    for v_name, pblock in pre_assignments_in_str.items() 
  }

  def _get_dir(dir_in_str: str) -> Dir:
    if dir_in_str == 'HORIZONTAL':
      return Dir.horizontal
    elif dir_in_str == 'VERTICAL':
      return Dir.vertical
    else:
      _logger.error(f'unknown partition direction {dir_in_str}')
      assert False

  partition_order = [_get_dir(dir_in_str) for dir_in_str in partition_order_in_str]

  print_pre_assignment(pre_assignments)

  print_vertex_areas(init_v2s.keys())

  # choose floorplan method
  num_vertices = len(graph.getAllVertices())
  v2s: Dict[Vertex, Slot] = {}
  if num_vertices < threshold_for_iterative:
    _logger.info(f'There are {num_vertices} vertices in the design, use eight way partition')

    if num_vertices > 100:
      _logger.warning('Over 100 vertices. May have a long solving time. Reduce threshold_for_iterative to skip to iterative bi-partitioning.')

    v2s = eight_way_partition(init_v2s, slot_manager, grouping_constraints, pre_assignments, ref_usage_ratio)
    if v2s:
      return v2s
    else:
      _logger.warning(f'Please check if any function in the design is too large')
  
  _logger.info(f'Use four-way partition because eight-way partition failed or there are too many vertices ({num_vertices})')
  v2s = four_way_partition(init_v2s, slot_manager, grouping_constraints, pre_assignments, ref_usage_ratio)
  if v2s:
    return v2s

  v2s = iterative_bipartition(init_v2s, slot_manager, grouping_constraints, pre_assignments, partition_order, ref_usage_ratio)
  if v2s:
    return v2s

  _logger.error(f'AutoBridge fails to partition the design at the SLR level. Either the design is too large, or the functions/modules are too large.')
  return None



fifo_template = '''

// first-word fall-through (FWFT) FIFO that is friendly for floorplanning
module relay_station #(
    parameter DATA_WIDTH = 32,
    parameter ADDR_WIDTH = 5,
    parameter DEPTH      = 2,
    parameter LEVEL      = 2
) (
  input wire clk,
  input wire reset,

  // write
  output wire                  if_full_n,
  input  wire                  if_write_ce,
  input  wire                  if_write,
  input  wire [DATA_WIDTH-1:0] if_din,

  // read
  output wire                  if_empty_n,
  input  wire                  if_read_ce,
  input  wire                  if_read,
  output wire [DATA_WIDTH-1:0] if_dout
);

  (* dont_touch = "yes" *) wire                  full_n  [LEVEL:0];
  (* dont_touch = "yes" *) wire                  empty_n [LEVEL:0];
  (* dont_touch = "yes" *) wire [DATA_WIDTH-1:0] data    [LEVEL:0];

  // both full_n and write are registered, thus one level of relay_station cause two additional latency for the almost full fifo
  parameter GRACE_PERIOD = LEVEL * 2;
  parameter REAL_DEPTH = GRACE_PERIOD + DEPTH + 4;
  parameter REAL_ADDR_WIDTH  = $clog2(REAL_DEPTH);

  genvar i;
  for (i = 0; i < LEVEL; i = i + 1) begin : inst
    if (i < LEVEL - 1) begin
      fifo_reg #(
        .DATA_WIDTH(DATA_WIDTH)
      ) reg_unit (
        .clk(clk),
        .reset(reset),

        // connect to fifo[i+1]
        .if_empty_n(empty_n[i+1]),
        .if_read_ce(if_read_ce),
        .if_read   (full_n[i+1]),
        .if_dout   (data[i+1]),

        // connect to fifo[i-1]
        .if_full_n  (full_n[i]),
        .if_write_ce(if_write_ce),
        .if_write   (empty_n[i]),
        .if_din     (data[i])
      );

    end else begin
      (* keep = "true" *) fifo_almost_full #(
        .DATA_WIDTH(DATA_WIDTH),
        .ADDR_WIDTH(REAL_ADDR_WIDTH),
        .DEPTH(REAL_DEPTH),
        .GRACE_PERIOD(GRACE_PERIOD)
      ) fifo_unit (
        .clk(clk),
        .reset(reset),

        // connect to fifo[i+1]
        .if_empty_n(empty_n[i+1]),
        .if_read_ce(if_read_ce),
        .if_read   (full_n[i+1]),
        .if_dout   (data[i+1]),

        // connect to fifo[i-1]
        .if_full_n  (full_n[i]),
        .if_write_ce(if_write_ce),
        .if_write   (empty_n[i]),
        .if_din     (data[i])
      );
    end
  end

  // write
  assign if_full_n  = full_n[0];  // output
  assign empty_n[0] = if_write;   // input
  assign data[0]    = if_din;     // input

  // read
  assign if_empty_n    = empty_n[LEVEL];  // output
  assign full_n[LEVEL] = if_read;         // input
  assign if_dout       = data[LEVEL];     // output

endmodule   // relay_station

/////////////////////////////////////////////////////////////

module fifo_reg #(
  parameter DATA_WIDTH = 32
) (
  input wire clk,
  input wire reset,

  // write
  (* dont_touch = "yes" *) output reg                  if_full_n,
  input  wire                  if_write_ce,
  input  wire                  if_write,
  input  wire [DATA_WIDTH-1:0] if_din,

  // read
  (* dont_touch = "yes" *) output reg                  if_empty_n,
  input  wire                  if_read_ce,
  input  wire                  if_read,
  (* dont_touch = "yes" *) output reg [DATA_WIDTH-1:0] if_dout
);

  always @ (posedge clk) begin
    if_dout <= if_din;
    if_empty_n <= if_write;
    if_full_n <= if_read;
  end

endmodule

/////////////////////////////////////////////////////////////////
// first-word fall-through (FWFT) FIFO
module fifo_almost_full #(
  parameter DATA_WIDTH = 32,
  parameter ADDR_WIDTH = 5,
  parameter DEPTH      = 32,
  parameter GRACE_PERIOD = 2,
  parameter USE_BRAM = 0,
  parameter USE_SRL = 0
) (
  input wire clk,
  input wire reset,
  // write
  output wire                  if_full_n,
  input  wire                  if_write_ce,
  input  wire                  if_write,
  input  wire [DATA_WIDTH-1:0] if_din,
  // read
  output wire                  if_empty_n,
  input  wire                  if_read_ce,
  input  wire                  if_read,
  output wire [DATA_WIDTH-1:0] if_dout
);
  parameter REAL_DEPTH = GRACE_PERIOD + DEPTH + 4;
  parameter REAL_ADDR_WIDTH  = $clog2(REAL_DEPTH);
generate
  if ( USE_BRAM != 0 && USE_SRL == 0) begin : bram
    fifo_bram_almost_full #(
      .DATA_WIDTH(DATA_WIDTH),
      .ADDR_WIDTH(REAL_ADDR_WIDTH),
      .DEPTH     (REAL_DEPTH),
      .GRACE_PERIOD(GRACE_PERIOD) /*********/
    ) unit (
      .clk  (clk),
      .reset(reset),
      .if_full_n  (if_full_n),
      .if_write_ce(if_write_ce),
      .if_write   (if_write),
      .if_din     (if_din),
      .if_empty_n(if_empty_n),
      .if_read_ce(if_read_ce),
      .if_read   (if_read),
      .if_dout   (if_dout)
    );
  end 
  else if ( USE_SRL != 0 && USE_BRAM == 0 ) begin : srl
    fifo_srl_almost_full #(
      .DATA_WIDTH(DATA_WIDTH),
      .ADDR_WIDTH(REAL_ADDR_WIDTH),
      .DEPTH     (REAL_DEPTH),
      .GRACE_PERIOD(GRACE_PERIOD) /*********/
    ) unit (
      .clk  (clk),
      .reset(reset),
      .if_full_n  (if_full_n),
      .if_write_ce(if_write_ce),
      .if_write   (if_write),
      .if_din     (if_din),
      .if_empty_n(if_empty_n),
      .if_read_ce(if_read_ce),
      .if_read   (if_read),
      .if_dout   (if_dout)
    );
  end
  else if ( DEPTH >= 1024 ) begin : bram
    fifo_bram_almost_full #(
      .DATA_WIDTH(DATA_WIDTH),
      .ADDR_WIDTH(REAL_ADDR_WIDTH),
      .DEPTH     (REAL_DEPTH),
      .GRACE_PERIOD(GRACE_PERIOD) /*********/
    ) unit (
      .clk  (clk),
      .reset(reset),
      .if_full_n  (if_full_n),
      .if_write_ce(if_write_ce),
      .if_write   (if_write),
      .if_din     (if_din),
      .if_empty_n(if_empty_n),
      .if_read_ce(if_read_ce),
      .if_read   (if_read),
      .if_dout   (if_dout)
    );
  end else begin : srl
    fifo_srl_almost_full #(
      .DATA_WIDTH(DATA_WIDTH),
      .ADDR_WIDTH(REAL_ADDR_WIDTH),
      .DEPTH     (REAL_DEPTH),
      .GRACE_PERIOD(GRACE_PERIOD) /*********/
    ) unit (
      .clk  (clk),
      .reset(reset),
      .if_full_n  (if_full_n),
      .if_write_ce(if_write_ce),
      .if_write   (if_write),
      .if_din     (if_din),
      .if_empty_n(if_empty_n),
      .if_read_ce(if_read_ce),
      .if_read   (if_read),
      .if_dout   (if_dout)
    );
  end
endgenerate
endmodule  // fifo
/////////////////////////////////////////////////////////////////
module fifo_srl_almost_full (
    clk,
    reset,
    if_empty_n,
    if_read_ce,
    if_read,
    if_dout,
    if_full_n,
    if_write_ce,
    if_write,
    if_din);
parameter MEM_STYLE   = "shiftreg";
parameter DATA_WIDTH  = 32'd32;
parameter ADDR_WIDTH  = 32'd4;
parameter DEPTH       = 5'd16;
/*******************************************/
parameter GRACE_PERIOD = 2;
/*******************************************/
input clk;
input reset;
output if_empty_n;
input if_read_ce;
input if_read;
output[DATA_WIDTH - 1:0] if_dout;
output if_full_n;
input if_write_ce;
input if_write;
input[DATA_WIDTH - 1:0] if_din;
wire[ADDR_WIDTH - 1:0] shiftReg_addr ;
wire[DATA_WIDTH - 1:0] shiftReg_data, shiftReg_q;
wire                     shiftReg_ce;
reg[ADDR_WIDTH:0] mOutPtr = ~{(ADDR_WIDTH+1){1'b0}};
reg internal_empty_n = 0, internal_full_n = 1;
assign if_empty_n = internal_empty_n;
/*******************************************/
// assign if_full_n = internal_full_n;
wire almost_full = mOutPtr >= DEPTH - 1 - GRACE_PERIOD && mOutPtr != ~{ADDR_WIDTH+1{1'b0}};
assign if_full_n = ~almost_full;
/*******************************************/
assign shiftReg_data = if_din;
assign if_dout = shiftReg_q;
always @ (posedge clk) begin
    if (reset == 1'b1)
    begin
        mOutPtr <= ~{ADDR_WIDTH+1{1'b0}};
        internal_empty_n <= 1'b0;
        internal_full_n <= 1'b1;
    end
    else begin
        if (((if_read & if_read_ce) == 1 & internal_empty_n == 1) && 
            ((if_write & if_write_ce) == 0 | internal_full_n == 0))
        begin
            mOutPtr <= mOutPtr - 5'd1;
            if (mOutPtr == 0)
                internal_empty_n <= 1'b0;
            internal_full_n <= 1'b1;
        end 
        else if (((if_read & if_read_ce) == 0 | internal_empty_n == 0) && 
            ((if_write & if_write_ce) == 1 & internal_full_n == 1))
        begin
            mOutPtr <= mOutPtr + 5'd1;
            internal_empty_n <= 1'b1;
            if (mOutPtr == DEPTH - 5'd2)
                internal_full_n <= 1'b0;
        end 
    end
end
assign shiftReg_addr = mOutPtr[ADDR_WIDTH] == 1'b0 ? mOutPtr[ADDR_WIDTH-1:0]:{ADDR_WIDTH{1'b0}};
assign shiftReg_ce = (if_write & if_write_ce) & internal_full_n;
fifo_srl_almost_full_internal 
#(
    .DATA_WIDTH(DATA_WIDTH),
    .ADDR_WIDTH(ADDR_WIDTH),
    .DEPTH(DEPTH))
U_fifo_w32_d16_A_ram (
    .clk(clk),
    .data(shiftReg_data),
    .ce(shiftReg_ce),
    .a(shiftReg_addr),
    .q(shiftReg_q));
endmodule  
module fifo_srl_almost_full_internal (
    clk,
    data,
    ce,
    a,
    q);
parameter DATA_WIDTH = 32'd32;
parameter ADDR_WIDTH = 32'd4;
parameter DEPTH = 5'd16;
input clk;
input [DATA_WIDTH-1:0] data;
input ce;
input [ADDR_WIDTH-1:0] a;
output [DATA_WIDTH-1:0] q;
reg[DATA_WIDTH-1:0] SRL_SIG [0:DEPTH-1];
integer i;
always @ (posedge clk)
    begin
        if (ce)
        begin
            for (i=0;i<DEPTH-1;i=i+1)
                SRL_SIG[i+1] <= SRL_SIG[i];
            SRL_SIG[0] <= data;
        end
    end
assign q = SRL_SIG[a];
endmodule
///////////////////////////////////////////////////////////
// first-word fall-through (FWFT) FIFO using block RAM or URAM (let Vivado choose)
// based on HLS generated code
module fifo_bram_almost_full #(
  parameter MEM_STYLE  = "auto",
  parameter DATA_WIDTH = 32,
  parameter ADDR_WIDTH = 5,
  parameter DEPTH      = 32,
  parameter GRACE_PERIOD = 2
) (
  input wire clk,
  input wire reset,
  // write
  output wire                  if_full_n,
  input  wire                  if_write_ce,
  input  wire                  if_write,
  input  wire [DATA_WIDTH-1:0] if_din,
  // read
  output wire                  if_empty_n,
  input  wire                  if_read_ce,
  input  wire                  if_read,
  output wire [DATA_WIDTH-1:0] if_dout
);
(* ram_style = MEM_STYLE *)
reg  [DATA_WIDTH-1:0] mem[0:DEPTH-1];
reg  [DATA_WIDTH-1:0] q_buf;
reg  [ADDR_WIDTH-1:0] waddr;
reg  [ADDR_WIDTH-1:0] raddr;
wire [ADDR_WIDTH-1:0] wnext;
wire [ADDR_WIDTH-1:0] rnext;
wire                  push;
wire                  pop;
reg  [ADDR_WIDTH-1:0] used;
reg                   full_n;
reg                   empty_n;
reg  [DATA_WIDTH-1:0] q_tmp;
reg                   show_ahead;
reg  [DATA_WIDTH-1:0] dout_buf;
reg                   dout_valid;
localparam DepthM1 = DEPTH[ADDR_WIDTH-1:0] - 1'd1;
/**************************************/
wire almost_full = (used >= DEPTH - 1 - GRACE_PERIOD);
//assign if_full_n  = full_n;
assign if_full_n  = ~almost_full;
/**************************************/
assign if_empty_n = dout_valid;
assign if_dout    = dout_buf;
assign push       = full_n & if_write_ce & if_write;
assign pop        = empty_n & if_read_ce & (~dout_valid | if_read);
assign wnext      = !push              ? waddr              :
                    (waddr == DepthM1) ? {ADDR_WIDTH{1'b0}} : waddr + 1'd1;
assign rnext      = !pop               ? raddr              :
                    (raddr == DepthM1) ? {ADDR_WIDTH{1'b0}} : raddr + 1'd1;
// waddr
always @(posedge clk) begin
  if (reset)
    waddr <= {ADDR_WIDTH{1'b0}};
  else
    waddr <= wnext;
end
// raddr
always @(posedge clk) begin
  if (reset)
    raddr <= {ADDR_WIDTH{1'b0}};
  else
    raddr <= rnext;
end
// used
always @(posedge clk) begin
  if (reset)
    used <= {ADDR_WIDTH{1'b0}};
  else if (push && !pop)
    used <= used + 1'b1;
  else if (!push && pop)
    used <= used - 1'b1;
end
// full_n
always @(posedge clk) begin
  if (reset)
    full_n <= 1'b1;
  else if (push && !pop)
    full_n <= (used != DepthM1);
  else if (!push && pop)
    full_n <= 1'b1;
end
// empty_n
always @(posedge clk) begin
  if (reset)
    empty_n <= 1'b0;
  else if (push && !pop)
    empty_n <= 1'b1;
  else if (!push && pop)
    empty_n <= (used != {{(ADDR_WIDTH-1){1'b0}},1'b1});
end
// mem
always @(posedge clk) begin
  if (push)
    mem[waddr] <= if_din;
end
// q_buf
always @(posedge clk) begin
  q_buf <= mem[rnext];
end
// q_tmp
always @(posedge clk) begin
  if (reset)
    q_tmp <= {DATA_WIDTH{1'b0}};
  else if (push)
    q_tmp <= if_din;
end
// show_ahead
always @(posedge clk) begin
  if (reset)
    show_ahead <= 1'b0;
  else if (push && used == {{(ADDR_WIDTH-1){1'b0}},pop})
    show_ahead <= 1'b1;
  else
    show_ahead <= 1'b0;
end
// dout_buf
always @(posedge clk) begin
  if (reset)
    dout_buf <= {DATA_WIDTH{1'b0}};
  else if (pop)
    dout_buf <= show_ahead? q_tmp : q_buf;
end
// dout_valid
always @(posedge clk) begin
  if (reset)
    dout_valid <= 1'b0;
  else if (pop)
    dout_valid <= 1'b1;
  else if (if_read_ce & if_read)
    dout_valid <= 1'b0;
end
endmodule  // fifo_bram

'''


import logging

from collections import defaultdict
from typing import Dict, List, Set
from autobridge.Opt.DataflowGraph import Vertex, Edge
from autobridge.Opt.Slot import Slot

RESOURCE_TYPES = ['BRAM', 'DSP', 'FF', 'LUT', 'URAM']

_logger = logging.getLogger().getChild(__name__)


def invert_v2s(curr_v2s: Dict[Vertex, Slot]) -> Dict[Slot, List[Vertex]]:
  s2v = defaultdict(list)
  for v, s in curr_v2s.items():
    s2v[s].append(v)
  return dict(s2v)


def get_all_edges(v_group: List[Vertex]) -> List[Edge]:
  """
  """
  return list(set([e for v in v_group for e in v.getEdges()]))


def get_inter_edges(v_group: List[Vertex]) -> List[Edge]:
  """
  get edges such that either src or sink is among v_group
  """
  first_visit: Set[Edge] = set()

  # if an edge is visited twice, then it is entirely within the target slots
  # if an edge is visited only once, then it is between the target slots and the remaining slots
  for e in get_all_edges(v_group):
    if e in first_visit:
      first_visit.remove(e)
    else:
      first_visit.add(e)

  inter_edges = list(first_visit)

  return inter_edges


def get_intra_edges(v_group: List[Vertex]) -> List[Edge]:
  """
  get edges whose src and sink are all among v_group
  """
  inter_edges = get_inter_edges(v_group)
  return [e for e in get_all_edges(v_group) if e not in inter_edges]


def get_total_wirelength(
    v2s: Dict[Vertex, Slot], 
) -> int:
  e_list = get_all_edges(list(v2s.keys()))
  wire_length_list = []
  for e in e_list:
    src_slot = v2s[e.src]
    dst_slot = v2s[e.dst]
    length = abs(src_slot.getPositionY() - dst_slot.getPositionY()) + \
              abs(src_slot.getPositionX() - dst_slot.getPositionX())
    wire_length_list.append(length * e.width)
  
  return sum(wire_length_list)


def log_resource_utilization(
    v2s: Dict[Vertex, Slot], 
) -> None:
  """
  analyze and log the new floorplan results
  """
  s2v = invert_v2s(v2s)
  for s, v_list in s2v.items():
    _logger.info(f'Slot {s.getRTLModuleName()}:')
    for r in RESOURCE_TYPES:
      capacity = s.area[r]
      usage = sum(v.area[r] for v in v_list)
      _logger.info(f'  [{r.ljust(4, " ")}]: {round(usage/capacity*100, 1)}% ({usage} / {capacity})')

  for s, v_list in s2v.items():
    _logger.info(f'Slot {s.getRTLModuleName()} contains:')
    for v in v_list:
      area_info = ' '.join(f'{r}: {v.area[r]}' for r in RESOURCE_TYPES)
      _logger.info(f'  {v.name}: {area_info}')

  # wire information
  _logger.info(f'total wire length: {get_total_wirelength(v2s)}')

  # SLR crossing information
  try:
    slr_crossing = [0] * (_get_slr_count(list(s2v.keys())) - 1)
    e_list = get_all_edges(list(v2s.keys()))
    for e in e_list:
      src_slr = v2s[e.src].getSLR()
      dst_slr = v2s[e.dst].getSLR()
      idx_small = min(src_slr, dst_slr)
      idx_large = max(src_slr, dst_slr)
      for i in range(idx_small, idx_large):
        slr_crossing[i] += e.width

    for i, count in enumerate(slr_crossing):
      _logger.info(f'SLR boundary {i} - {i+1} has {slr_crossing[i]} crossings')
  except:
    _logger.info(f'Skip logging SLR crossing because current slots include multiple SLRs')


def _get_slr_count(slot_list: List[Slot]) -> int:
  """
  how many SLR does the board has
  """
  slr_count_list = [slot.board.SLR_NUM for slot in slot_list]
  assert all(slr_count_list[i] == slr_count_list[0] for i in range(len(slr_count_list)))
  return slr_count_list[0]


def print_vertex_areas(v_list: List[Vertex]) -> None:
  _logger.info('The area of each vertex is listed as below.')
  _logger.info('Note that the area of each vertex includes the area of all its in-bound FIFOs.')
  for v in v_list:
    _logger.info(f'    {v.name}: ' + ' '.join(f'{r}: {v.getVertexAndInboundFIFOArea()[r]}' for r in RESOURCE_TYPES))


def print_pre_assignment(v2s: Dict[Vertex, Slot]) -> None:
  _logger.info('The pre-determined floorplanning is shown as below:')
  for v, s in v2s.items():
    _logger.info(f'    {v.name}  =>  {s.getRTLModuleName()}')


def float_range(start, stop, step) -> List[float]:
  range = []
  while start <= stop:
    range.append(start)
    start = round(start + step, 2)  
  return range


from typing import Dict, List

from autobridge.Opt.DataflowGraph import Vertex, Edge
from autobridge.Opt.GlobalRouting import GlobalRouting
from autobridge.Opt.Slot import Slot
from autobridge.Opt.Floorplan import Floorplanner
from autobridge.Device.DeviceManager import DeviceBase

VITIS_HIERARCHY_ADDRESS = 'pfm_top_i/dynamic_region/.*/inst'


def get_used_slots(floorplan, global_router) -> List[Slot]:
  """
  (1) at least one vertex is assigned
  or (2) at least one pipeline registers is assigned 
  """
  vertex_slot = list(floorplan.getSlotToVertices().keys())
  
  pipeline_reg_slot = []
  for e_list in floorplan.getSlotToEdges().values():
    for e in e_list:
      if e.pipeline_level == 0:
        continue

      pipeline_reg_slot += global_router.e_name2path[e.name]  # exclude src and dst

  all_slot_list = list(set(vertex_slot + pipeline_reg_slot))

  return all_slot_list
  

def create_pblocks(slot_list: List[Slot], board: DeviceBase) -> List[str]:
  tcl = []
  for slot in slot_list:
    tcl += board.getSlotPblockTcl(slot)
    
  return tcl


def gen_constraints_for_vertices(s2v: Dict[Slot, List[Vertex]]) -> List[str]:
  tcl = []
  for slot, v_list in s2v.items():
    assert v_list
    tcl.append(f'add_cells_to_pblock {slot.getRTLModuleName()} [ get_cells -regexp {{ ')
    for v in v_list:
      tcl.append(f'  {VITIS_HIERARCHY_ADDRESS}/{v.name}')
    tcl.append(f'}} ]')

  return tcl


def gen_constraints_for_almost_full_fifos(s2e: Dict[Slot, List[Edge]]) -> List[str]:
  tcl = []
  for slot, e_list in s2e.items():
    almost_full_fifo_edges = [e for e in e_list if e.pipeline_level == 0]
    if almost_full_fifo_edges:
      tcl.append(f'add_cells_to_pblock {slot.getRTLModuleName()} [ get_cells -regexp {{ ')
      for e in almost_full_fifo_edges:
        tcl.append(f'  {VITIS_HIERARCHY_ADDRESS}/{e.name}')
      tcl.append(f'}} ]')

  return tcl


def gen_constraints_for_relay_stations(
    s2e: Dict[Slot, List[Edge]],
    v2s: Dict[Vertex, Slot],
    global_router: GlobalRouting,
) -> List[str]:
  """
  assign each pipeline stage of the relay station to one slot
  """
  tcl = []
  for e_list in s2e.values():
    for e in e_list:
      if e.pipeline_level == 0:
        continue

      slot_path = global_router.e_name2path[e.name]  # exclude src and dst
      slot_path_with_src_and_dst = [ v2s[e.src], *slot_path, v2s[e.dst] ]

      slot_crossing_num = len(slot_path) + 1
      assert e.pipeline_level == slot_crossing_num * 2, 'check if the pipeline policy has changed'

      for i in range(slot_crossing_num):
        tcl.append(f'add_cells_to_pblock {slot_path_with_src_and_dst[i].getRTLModuleName()} [get_cells -regexp {{ {VITIS_HIERARCHY_ADDRESS}/{e.name}/inst.*{2*i}.*unit }}]')
        tcl.append(f'add_cells_to_pblock {slot_path_with_src_and_dst[i+1].getRTLModuleName()} [get_cells -regexp {{ {VITIS_HIERARCHY_ADDRESS}/{e.name}/inst.*{2*i+1}.*unit }}]')

      # always constrain that the final fifo_unit is at the same slot with the dst vertex
      # should be redundant
      tcl.append(f'add_cells_to_pblock {slot_path_with_src_and_dst[-1].getRTLModuleName()} [get_cells -regexp {{ {VITIS_HIERARCHY_ADDRESS}/{e.name}/inst.*fifo_unit }}]')

  return tcl


def generate_floorplan_constraints(floorplan: Floorplanner, global_router: GlobalRouting):
  tcl = []

  tcl += ['write_checkpoint before_applying_floorplan_constriants.dcp']

  slot_list = get_used_slots(floorplan, global_router)
  
  tcl += create_pblocks(slot_list, floorplan.board)

  tcl += gen_constraints_for_vertices(floorplan.getSlotToVertices())

  tcl += gen_constraints_for_almost_full_fifos(floorplan.getSlotToEdges())

  tcl += gen_constraints_for_relay_stations(floorplan.getSlotToEdges(), floorplan.getVertexToSlot(), global_router)

  open('constraint.tcl', 'w').write('\n'.join(tcl))


from mip import Model, Var, BINARY, xsum

from autobridge.Floorplan.Utilities import *
from autobridge.Opt.DataflowGraph import Vertex, Edge

# refer to https://cs.stackexchange.com/questions/12102/express-boolean-logic-operations-in-zero-one-integer-linear-programming-ilp

def get_var_of_logic_and(m: Model, x1: Var, x2: Var) -> Var:
  """ return a binary variable represents the logic AND """
  _check_if_vars_binary(x1, x2)

  y = m.add_var(var_type=BINARY)
  m += y >= x1 + x2 - 1
  m += y <= x1
  m += y <= x2
  
  return y


def get_var_of_logic_or(m: Model, x1: Var, x2: Var) -> Var:
  """ return a binary variable represents the logic OR """
  _check_if_vars_binary(x1, x2)

  y = m.add_var(var_type=BINARY)
  m += y <= x1 + x2
  m += y >= x1
  m += y >= x2
  
  return y


def get_var_of_logic_xor(m: Model, x1: Var, x2: Var) -> Var:
  """ return a binary variable represents the logic XOR """
  _check_if_vars_binary(x1, x2)

  y = m.add_var(var_type=BINARY)
  m += y <= x1 + x2
  m += y >= x1 - x2
  m += y >= x2 - x1
  m += y <= 2 - x1 - x2

  return y


def get_var_of_logic_not(m: Model, x: Var) -> Var:
  """ return a binary variable represents the logic NOT """
  _check_if_vars_binary(x)

  y = m.add_var(var_type=BINARY)
  m += y >= 1 - x
  m += y <= 1 - x
  
  return y


def get_var_of_equal_zero(m: Model, x: Var) -> Var:
  """ return a binary variable represents if the input binary var equals zero """
  _check_if_vars_binary(x)
  return get_var_of_logic_not(m, x)

def get_var_of_equal_one(m: Model, x: Var) -> Var:
  _check_if_vars_binary(x)
  return x

def _check_if_vars_binary(*args) -> bool:
  """ check if the variables are BINARY """
  return all(v.var_type == BINARY for v in args)


def add_slr_0_1_crossing_constraint(
  m: Model,
  v_list: List[Vertex],
  v2var_y1: Dict[Vertex, Var],
  v2var_y2: Dict[Vertex, Var],
  width_limit: int,
) -> None:
  """ restrict the SLR crossing between 0 and 1 """
  all_edges = get_all_edges(v_list)

  def is_edge_cross_slr_0_1(e: Edge) -> Var:
    src_in_slr0 = get_var_of_logic_and(
      m, 
      get_var_of_equal_zero(m, v2var_y1[e.src]),
      get_var_of_equal_zero(m, v2var_y2[e.src])
    )
    dst_in_slr0 = get_var_of_logic_and(
      m, 
      get_var_of_equal_zero(m, v2var_y1[e.dst]),
      get_var_of_equal_zero(m, v2var_y2[e.dst])
    )
    src_not_in_slr0 = get_var_of_logic_not(m, src_in_slr0)
    dst_not_in_slr0 = get_var_of_logic_not(m, dst_in_slr0)

    return get_var_of_logic_or(
      m,
      get_var_of_logic_and(m, src_in_slr0, dst_not_in_slr0),
      get_var_of_logic_and(m, dst_in_slr0, src_not_in_slr0)
    )

  m += xsum(e.width * is_edge_cross_slr_0_1(e) for e in all_edges) <= width_limit 


def add_slr_1_2_crossing_constraint(
  m: Model,
  v_list: List[Vertex],
  v2var_y1: Dict[Vertex, Var],
  width_limit: int,
) -> None:
  """ restrict the SLR crossing between 1 and 2 """
  all_edges = get_all_edges(v_list)

  def is_edge_cross_slr_1_2(e: Edge) -> Var:
    return get_var_of_logic_xor(m, v2var_y1[e.src], v2var_y1[e.dst])
  
  m += xsum(e.width * is_edge_cross_slr_1_2(e) for e in all_edges) <= width_limit 


def add_slr_2_3_crossing_constraint(
  m: Model,
  v_list: List[Vertex],
  v2var_y1: Dict[Vertex, Var],
  v2var_y2: Dict[Vertex, Var],
  width_limit: int,
) -> None:
  """ restrict the SLR crossing between 2 and 3 """
  all_edges = get_all_edges(v_list)

  def is_edge_cross_slr_2_3(e: Edge) -> Var:
    src_in_slr3 = get_var_of_logic_and(
      m, 
      get_var_of_equal_one(m, v2var_y1[e.src]),
      get_var_of_equal_one(m, v2var_y2[e.src])
    )
    dst_in_slr3 = get_var_of_logic_and(
      m, 
      get_var_of_equal_one(m, v2var_y1[e.dst]),
      get_var_of_equal_one(m, v2var_y2[e.dst])
    )
    src_not_in_slr3 = get_var_of_logic_not(m, src_in_slr3)
    dst_not_in_slr3 = get_var_of_logic_not(m, dst_in_slr3)

    return get_var_of_logic_or(
      m,
      get_var_of_logic_and(m, src_in_slr3, dst_not_in_slr3),
      get_var_of_logic_and(m, dst_in_slr3, src_not_in_slr3)
    )

  m += xsum(e.width * is_edge_cross_slr_2_3(e) for e in all_edges) <= width_limit



import logging
from mip import *

class LatencyBalancing:
  def __init__(self, graph, floorplan, global_router):
    logging.info('running latency balancing...')
    self.name2edge = graph.getNameToEdgeMap()
    self.name2vertex = graph.getNameToVertexMap()
    self.floorplan = floorplan
    self.global_router = global_router

    self.__rebalance()

  def __rebalance(self):
    m = Model()

    # map Vertex -> "arrival time"
    v2var = {}
    for name, v in self.name2vertex.items():
      v2var[v] = m.add_var(var_type=INTEGER, name=name)

    # differential constraint for each edge
    for e_name, e in self.name2edge.items():
      # +1 because each FIFO by itself has 1 unit of latency
      # in case the original design is not balanced
      # note that additional pipelining for full_n will not lead to additional latency
      # we only need to increase the grace period of almost full FIFOs by 1
      # [update]: we skip the +1 for the orginial FIFO. We only take care of our own modifications
      m += v2var[e.src] >= v2var[e.dst] + e.pipeline_level

    m.objective = minimize(xsum( 
      e.width * (v2var[e.src] - v2var[e.dst]) for e in self.name2edge.values() 
    ))

    status = m.optimize(max_seconds=120)
    assert status == OptimizationStatus.OPTIMAL or status == OptimizationStatus.FEASIBLE, 'Failed to balance reconvergent paths at loop level. Dependency loop detected.'

    # get result
    for e_name, e in self.name2edge.items():
      e.added_depth_for_rebalance = int(v2var[e.src].x - v2var[e.dst].x) - e.pipeline_level
      assert e.added_depth_for_rebalance >= 0

    # logging
    for e_name, e in self.name2edge.items():
      logging.info(f'{e_name}: pipeline_level: {e.pipeline_level}, added_depth_for_rebalance: {e.added_depth_for_rebalance}, width: {e.width} ')



import enum
import re
import math

from typing import List
from autobridge.Opt.Slot import Slot


# partition directions
class Dir(enum.Enum):
  horizontal = 1
  vertical = 2


class SlotManager:
  def __init__(self, board):
    self.board = board
    self.pblock_to_slot = {}
    self.pblock_to_routing_slot = {}

  def _preprocessPblock(self, pblock : str) -> str:
    def _convertCoarseRegionToClockRegion(coarse_loc):
      match = re.search(r'COARSE_X(\d+)Y(\d+)', coarse_loc)
      assert match

      x = int(match.group(1))
      y = int(match.group(2))
      CR_NUM_HORIZONTAL_PER_COARSE = int(0.5 * self.board.CR_NUM_HORIZONTAL)
      CR_NUM_VERTICAL_PER_COARSE = int(self.board.CR_NUM_VERTICAL_PER_SLR)

      left_bottom_x = x * CR_NUM_HORIZONTAL_PER_COARSE
      left_bottom_y = y * CR_NUM_VERTICAL_PER_COARSE
      right_up_x = left_bottom_x + CR_NUM_HORIZONTAL_PER_COARSE - 1
      right_up_y = left_bottom_y + CR_NUM_VERTICAL_PER_COARSE - 1
      return f'CLOCKREGION_X{left_bottom_x}Y{left_bottom_y}:CLOCKREGION_X{right_up_x}Y{right_up_y}'

    if 'COARSE' in pblock:
      return _convertCoarseRegionToClockRegion(pblock)
    else:
      match = re.search(r'^CLOCKREGION_X(\d+)Y(\d+)[ ]*:[ ]*CLOCKREGION_X(\d+)Y(\d+)$', pblock)

      if not match:
        match = re.search(r'^CR_X(\d+)Y(\d+)[ ]*_To_[ ]*CR_X(\d+)Y(\d+)$', pblock)

      assert match, f'incorrect pblock {pblock}'
      return pblock

  def getAllTwoByTwoCRSlots(self):
    """
    [FIXME] a temporary fix. Get all Slots of 2x2 clock regions for U250
    """
    slot_list = []
    for x in range(0, 8, 2):
      for y in range(0, 16, 2):
        slot_list.append(Slot(self.board, f'CLOCKREGION_X{x}Y{y}:CLOCKREGION_X{x+1}Y{y+1}'))
    return slot_list

  def getLeafSlotsAfterPartition(
    self,
    partition_order: List[Dir],
  ):
    """
    generate all the final slots after a set of partition operations
    """
    init_slot = self.getInitialSlot()
    queue = [init_slot]
    for part_dir in partition_order:
      next_queue = []
      for slot in queue:
        if part_dir == Dir.horizontal:
          next_queue += self.getBottomAndUpSplit(slot)
        else:
          next_queue += self.getLeftAndRightSplit(slot)
      queue = next_queue

    return queue

  def createSlotForRouting(self, pblock : str):
    """create a Slot object for global routing purpose"""
    pblock = self._preprocessPblock(pblock)
    if pblock not in self.pblock_to_routing_slot:
      self.pblock_to_routing_slot[pblock] = Slot(self.board, pblock)
    return self.pblock_to_routing_slot[pblock]

  def createSlot(self, pblock : str):
    """create a Slot for floorplanning purpose and track the existing Slot objects"""
    pblock = self._preprocessPblock(pblock)
    if pblock not in self.pblock_to_slot:
      self.pblock_to_slot[pblock] = Slot(self.board, pblock)
    return self.pblock_to_slot[pblock]

  def getPureRoutingSlots(self):
    return [self.pblock_to_routing_slot[pblock] \
        for pblock in self.pblock_to_routing_slot.keys() \
        if pblock not in self.pblock_to_slot]

  def getComputeSlots(self):
    return self.pblock_to_slot.values()

  def removeSlotNonBlocking(self, pblock : str):
    if pblock in self.pblock_to_slot:
      self.pblock_to_slot.pop(pblock)

  # split by the middle row
  def getBottomAndUpSplit(self, slot):
    up     = self.createSlot(slot.getUpChildSlotName())
    bottom = self.createSlot(slot.getBottomChildSlotName())
    self.removeSlotNonBlocking(slot.getName())

    return bottom, up 

  # split by the middle column
  def getLeftAndRightSplit(self, slot):
    left =  self.createSlot(slot.getLeftChildSlotName())
    right = self.createSlot(slot.getRightChildSlotName())
    self.removeSlotNonBlocking(slot.getName())

    return left, right

  def partitionSlotByHalf(self, slot : Slot, dir : Dir):
    if dir == Dir.horizontal:
      return self.getBottomAndUpSplit(slot)
    elif dir == Dir.vertical:
      return self.getLeftAndRightSplit(slot)
    else:
      assert False, f'unrecognized partition direction: {dir}'

  def getInitialSlot(self):
    """
    get the initial slot for partitioning.
    Note that we expand the slot to the power of 2.
    This helps the partition process.
    The parts beyond the physical device will have 0 area
    """
    ceil_to_power_of_2 = lambda CR_NUM_VERTICAL : int(math.pow(2, math.ceil(math.log(CR_NUM_VERTICAL, 2))))
    UR_X = ceil_to_power_of_2(self.board.CR_NUM_HORIZONTAL) - 1
    UR_Y = ceil_to_power_of_2(self.board.CR_NUM_VERTICAL) - 1
    return self.createSlot(f'CLOCKREGION_X0Y0:CLOCKREGION_X{UR_X}Y{UR_Y}')

  def _haveOverlappedYRange(self, a : Slot, b : Slot):
    return min(a.getOrigUpRightY(), b.getOrigUpRightY()) > max(a.getOrigDownLeftY(), b.getOrigDownLeftY())
  
  def _haveOverlappedXRange(self, a : Slot, b : Slot):
    return min(a.getOrigUpRightX(), b.getOrigUpRightX()) > max(a.getOrigDownLeftX(), b.getOrigDownLeftX())
  
  def getLeftNeighborSlots(self, slot, potential_target_slots):
    neighbors = []
    for candidate in potential_target_slots.values():
      if candidate.getOrigUpRightX()+1 == slot.getOrigDownLeftX():
        if self._haveOverlappedYRange(candidate, slot):
          neighbors.append(candidate)
    return neighbors
  
  def getRightNeighborSlots(self, slot, potential_target_slots):
    neighbors = []
    for candidate in potential_target_slots.values():
      if slot.getOrigUpRightX()+1 == candidate.getOrigDownLeftX():
        if self._haveOverlappedYRange(candidate, slot):
          neighbors.append(candidate)
    return neighbors
  
  def getUpNeighborSlots(self, slot, potential_target_slots):
    neighbors = []
    for candidate in potential_target_slots.values():
      if slot.getOrigUpRightY()+1 == candidate.getOrigDownLeftY():
        if self._haveOverlappedXRange(candidate, slot):
          neighbors.append(candidate)
    return neighbors
  
  def getDownNeighborSlots(self, slot, potential_target_slots):
    neighbors = []
    for candidate in potential_target_slots.values():
      if candidate.getOrigUpRightY()+1 == slot.getOrigDownLeftY():
        if self._haveOverlappedXRange(candidate, slot):
          neighbors.append(candidate)
    return neighbors

  def getComputeNeighborSlots(self, slot, dir):
    potential_target_slots = self.pblock_to_slot

    if dir == 'UP':
      return self.getUpNeighborSlots(slot, potential_target_slots)
    elif dir == 'DOWN':
      return self.getDownNeighborSlots(slot, potential_target_slots)
    elif dir == 'LEFT':
      return self.getLeftNeighborSlots(slot, potential_target_slots)
    elif dir == 'RIGHT':
      return self.getRightNeighborSlots(slot, potential_target_slots)
    else:
      assert False, f'wrong direction: {dir}'

  def getActiveSlotsIncludeRouting(self):
    return {**self.pblock_to_slot, **self.pblock_to_routing_slot}.values()

  def isPureRoutingSlot(self, slot):
    slot_name = slot.getName()
    return slot_name in self.pblock_to_routing_slot and slot_name not in self.pblock_to_slot

  def getComputeSlotPairs(self):
    pairs = []
    for slot in self.pblock_to_slot.values():
      neighbors = self.getComputeNeighborSlots(slot, 'UP') + \
                  self.getComputeNeighborSlots(slot, 'RIGHT')
      pairs += [(slot, n) for n in neighbors]
    return pairs

  def getAnyNeighborSlots(self, slot, dir):
    potential_target_slots = {**self.pblock_to_slot, **self.pblock_to_routing_slot}

    if dir == 'UP':
      return self.getUpNeighborSlots(slot, potential_target_slots)
    elif dir == 'DOWN':
      return self.getDownNeighborSlots(slot, potential_target_slots)
    elif dir == 'LEFT':
      return self.getLeftNeighborSlots(slot, potential_target_slots)
    elif dir == 'RIGHT':
      return self.getRightNeighborSlots(slot, potential_target_slots)
    else:
      assert False, f'wrong direction: {dir}'

  def getAllSlotPairs(self):
    pairs = []
    for slot in self.getActiveSlotsIncludeRouting():
      neighbors = self.getAnyNeighborSlots(slot, 'UP') + \
                  self.getAnyNeighborSlots(slot, 'RIGHT')
      pairs += [(slot, n) for n in neighbors]
    return pairs


import re

from typing import List
from collections import defaultdict
from autobridge.Opt.Slot import Slot


class DeviceBase:
  NAME = 'Base'
  CR_AREA = None
  CR_NUM_VERTICAL = None
  FPGA_PART_NAME = None
  
  def __init__(self, ddr_list=[], is_vitis_enabled=True):
    self.ddr_list = ddr_list
    self.pre_existing_area = self._getVitisRegions(ddr_list, is_vitis_enabled)

  def _getCRPblockIntersect(self, cr_pblock1, cr_pblock2):
    """
    get the overlapped part of two pblocks of clock regions
    """
    assert re.search(r'CLOCKREGION_X\d+Y\d+:CLOCKREGION_X\d+Y\d+', cr_pblock1), cr_pblock1
    assert re.search(r'CLOCKREGION_X\d+Y\d+:CLOCKREGION_X\d+Y\d+', cr_pblock2), cr_pblock2
    pblock1_DL_x, pblock1_DL_y, pblock1_UR_x, pblock1_UR_y = \
      [int(val) for val in re.findall(r'[XY](\d+)', cr_pblock1)] # DownLeft & UpRight    
    pblock2_DL_x, pblock2_DL_y, pblock2_UR_x, pblock2_UR_y = \
      [int(val) for val in re.findall(r'[XY](\d+)', cr_pblock2)] # DownLeft & UpRight  
    
    intersect_DL_x = max(pblock1_DL_x, pblock2_DL_x)
    intersect_DL_y = max(pblock1_DL_y, pblock2_DL_y)
    intersect_UR_x = min(pblock1_UR_x, pblock2_UR_x)
    intersect_UR_y = min(pblock1_UR_y, pblock2_UR_y)

    if intersect_DL_x <= intersect_UR_x and intersect_DL_y <= intersect_UR_y:
      overlap_pblock = f'CLOCKREGION_X{intersect_DL_x}Y{intersect_DL_y}:CLOCKREGION_X{intersect_UR_x}Y{intersect_UR_y}'
    else:
      overlap_pblock = None

    return overlap_pblock

  def _getPblockArea(self, pblock_def):
    """
    get the total resources in the specified pblock
    Note that if a slot is outside the range of the device, we return an area of 0
    """
    assert re.search(r'CLOCKREGION_X\d+Y\d+:CLOCKREGION_X\d+Y\d+', pblock_def), f'unexpected format of the slot name {pblock_def}'
    DL_x, DL_y, UR_x, UR_y = [int(val) for val in re.findall(r'[XY](\d+)', pblock_def)] # DownLeft & UpRight

    # treat the pseudo SLR with 0 area
    UR_y = min(self.CR_NUM_VERTICAL-1, UR_y) 

    area = {
      'BRAM' : 0,
      'DSP' : 0,
      'FF' : 0,
      'LUT' : 0,
      'URAM' : 0
    }
    
    if DL_y > self.CR_NUM_VERTICAL-1:
      return area 

    for item in ['BRAM', 'DSP', 'FF', 'LUT', 'URAM']:
      # the total area of one row
      area[item] = sum(self.CR_AREA[i][j][item] for i in range(DL_x, UR_x + 1) for j in range(DL_y, UR_y+1))

    return area

  def getArea(self, slot_pblock):
    """
    get the resources available to user. Exclude any pre-exising IPs
    """
    slot_user_area = self._getPblockArea(slot_pblock)
    for ip_pblock in self.pre_existing_area:
      overlap_pblock = self._getCRPblockIntersect(slot_pblock, ip_pblock)
      if overlap_pblock:
        overlap_area = self._getPblockArea(overlap_pblock)
        for item in slot_user_area.keys():
          slot_user_area[item] -= overlap_area[item]
    return slot_user_area

  def getSlotPblockTcl(self, slot: Slot) -> List[str]:
    """
    remove the overlaps with vitis IPs
    If the slot is an SLR-level slot, then do not remove the DDR region in the middle
    """
    tcl = []
    pblock_name = slot.getRTLModuleName()
    pblock_def = slot.getName()

    tcl.append(f'create_pblock {pblock_name}')
    tcl.append(f'resize_pblock {pblock_name} -add {pblock_def}')
    tcl.append(f'# remove the reserved clock regions for the Vitis infra')
    tcl.append(f'resize_pblock {pblock_name} -remove {self.VITIS_REGION}')

    # exclude the DDR region if the DDR region is at the boundary of the slot
    for ddr in self.ddr_list:
      ddr_pblock = self.DDR_TO_CLOCK_REGIONS[ddr]
      ddr_slot = Slot(slot.board, ddr_pblock)
      if slot.down_left_x == ddr_slot.down_left_x or slot.up_right_x == ddr_slot.up_right_x:
        if slot.containsChildSlot(ddr_slot):
          if self.getDDRSlolenRegion(ddr):
            tcl.append(f'# exclude the area for DDR {ddr}')
            tcl.append(f'resize_pblock {pblock_name} -remove {self.DDR_TO_CLOCK_REGIONS[ddr]}')
            tcl.append(f'resize_pblock {pblock_name} -add {{ {self.getDDRSlolenRegion(ddr)} }}')

    return tcl

  def getDDRSlolenRegion(self, ddr: int) -> str:
    return ''

  DDR_TO_CLOCK_REGIONS = {}

  def getDDRPblock(self, ddr):
    return self.DDR_TO_CLOCK_REGIONS[ddr]


class DeviceU250(DeviceBase):
  def _getVitisRegions(self, ddr_list, is_vitis_enabled):
    assert all(ddr in [0, 1, 2, 3] for ddr in ddr_list) # u250 has 4 ddrs

    # the area used by implicit IPs
    pre_existing_area = []
    for ddr in ddr_list:
      pre_existing_area.append(self.DDR_TO_CLOCK_REGIONS[ddr])

    # the vitis platform will take away the rightmost column
    if is_vitis_enabled:
      pre_existing_area.append(self.VITIS_REGION) # the area consumed by Vitis platform
    
    return pre_existing_area

  def getDDRSlolenRegion(self, ddr: int) -> str:
    assert 0 <= ddr <= 3, ddr
    # entire sloten region is {SLICE_X144Y0:SLICE_X145Y959 DSP48E2_X19Y0:DSP48E2_X19Y383 RAMB18_X9Y0:RAMB18_X9Y383 RAMB36_X9Y0:RAMB36_X9Y191}
    return f'SLICE_X144Y{240 * ddr}:SLICE_X145Y{240 * (ddr+1) - 1} DSP48E2_X19Y{96 * ddr}:DSP48E2_X19Y{96 * (ddr+1) - 1} RAMB18_X9Y{96 * ddr}:RAMB18_X9Y{96 * (ddr+1) - 1} RAMB36_X9Y{48 * ddr}:RAMB36_X9Y{48 * (ddr+1) - 1}'

  NAME = 'U250'

  FPGA_PART_NAME = 'xcu250-figd2104-2L-e'
  # SLR level
  SLR_NUM = 4

  SLR_AREA = defaultdict(lambda: defaultdict(list))
  SLR_AREA['BRAM'][0] = 768
  SLR_AREA['DSP'][0] = 1536
  SLR_AREA['FF'][0] = 433920
  SLR_AREA['LUT'][0] = 216960

  SLR_AREA['BRAM'][1] = 384
  SLR_AREA['DSP'][1] = 1344
  SLR_AREA['FF'][1] = 329280
  SLR_AREA['LUT'][1] = 164640

  SLR_AREA_DDR = defaultdict(lambda: defaultdict(list))
  SLR_AREA_DDR['BRAM'][0] = 768
  SLR_AREA_DDR['DSP'][0] = 1536
  SLR_AREA_DDR['FF'][0] = 433920
  SLR_AREA_DDR['LUT'][0] = 216960

  SLR_AREA_DDR['BRAM'][1] = 288
  SLR_AREA_DDR['DSP'][1] = 1152
  SLR_AREA_DDR['FF'][1] = 245760
  SLR_AREA_DDR['LUT'][1] = 122800

  CR_NUM_HORIZONTAL = 8  
  CR_NUM_VERTICAL = 16  
  CR_NUM_VERTICAL_PER_SLR = 4 # each die has 4 CRs vertically
  
  # to be compatible with U280
  ACTUAL_SLR_NUM = 4
  ACTUAL_CR_NUM_VERTICAL = 16

  # Clock Region level
  CR_AREA = defaultdict(lambda: defaultdict(dict))
  for y in range(ACTUAL_CR_NUM_VERTICAL):
    # clock regions with Laguna columns have less FF and LUT
    if y % 4 == 0 or y % 3 == 0:
      LUT_adjustment = 0
      FF_adjustment = 0
    else:
      LUT_adjustment = 960
      FF_adjustment = 1920

    CR_AREA[0][y]['BRAM'] = 48
    CR_AREA[0][y]['DSP'] = 96
    CR_AREA[0][y]['FF'] = 27840 + FF_adjustment
    CR_AREA[0][y]['LUT'] = 13920 + LUT_adjustment
    CR_AREA[0][y]['URAM'] = 0

    CR_AREA[1][y]['BRAM'] = 48
    CR_AREA[1][y]['DSP'] = 96
    CR_AREA[1][y]['FF'] = 23040 + FF_adjustment
    CR_AREA[1][y]['LUT'] = 11520 + LUT_adjustment
    CR_AREA[1][y]['URAM'] = 16

    CR_AREA[2][y]['BRAM'] = 72
    CR_AREA[2][y]['DSP'] = 120
    CR_AREA[2][y]['FF'] = 34560 + FF_adjustment
    CR_AREA[2][y]['LUT'] = 17280 + LUT_adjustment
    CR_AREA[2][y]['URAM'] = 0

    CR_AREA[3][y]['BRAM'] = 24
    CR_AREA[3][y]['DSP'] = 72
    CR_AREA[3][y]['FF'] = 19200 + FF_adjustment
    CR_AREA[3][y]['LUT'] = 9600 + LUT_adjustment
    CR_AREA[3][y]['URAM'] = 16

    CR_AREA[4][y]['BRAM'] = 48
    CR_AREA[4][y]['DSP'] = 96
    CR_AREA[4][y]['FF'] = 25920 + FF_adjustment
    CR_AREA[4][y]['LUT'] = 12960 + LUT_adjustment
    CR_AREA[4][y]['URAM'] = 16

    CR_AREA[5][y]['BRAM'] = 24
    CR_AREA[5][y]['DSP'] = 120
    CR_AREA[5][y]['FF'] = 26880 + FF_adjustment
    CR_AREA[5][y]['LUT'] = 13440 + LUT_adjustment
    CR_AREA[5][y]['URAM'] = 16

    CR_AREA[6][y]['BRAM'] = 24
    CR_AREA[6][y]['DSP'] = 120
    CR_AREA[6][y]['FF'] = 26880 + FF_adjustment
    CR_AREA[6][y]['LUT'] = 13440 + LUT_adjustment
    CR_AREA[6][y]['URAM'] = 16

    CR_AREA[7][y]['BRAM'] = 48
    CR_AREA[7][y]['DSP'] = 48
    CR_AREA[7][y]['FF'] = 24000 + FF_adjustment
    CR_AREA[7][y]['LUT'] = 12000 + LUT_adjustment
    CR_AREA[7][y]['URAM'] = 0

  TOTAL_AREA = {}
  TOTAL_AREA['BRAM'] = 5376
  TOTAL_AREA['DSP'] = 12288
  TOTAL_AREA['FF'] = 3456000
  TOTAL_AREA['LUT'] = 1728000
  TOTAL_AREA['URAM'] = 1280

  # for item in ('BRAM', 'DSP', 'FF', 'LUT', 'URAM'):
  #   assert TOTAL_AREA[item] == sum(CR_AREA[x][y][item] for x in range(8) for y in range(16))

  # note that the location of DDR 3 is not consistent with other DDRs 
  DDR_TO_CLOCK_REGIONS = {
    0: 'CLOCKREGION_X4Y1:CLOCKREGION_X4Y3',
    1: 'CLOCKREGION_X4Y5:CLOCKREGION_X4Y7',
    2: 'CLOCKREGION_X4Y9:CLOCKREGION_X4Y11',
    3: 'CLOCKREGION_X4Y12:CLOCKREGION_X4Y14'
  }

  # the rightmost column
  VITIS_REGION = 'CLOCKREGION_X7Y0:CLOCKREGION_X7Y15'


class DeviceU280(DeviceBase):
  def _getVitisRegions(self, ddr_list, is_vitis_enabled):
    assert all(ddr in [0, 1] for ddr in ddr_list) # u280 only has 2 ddrs on SLR0 and SLR1

    # the area used by implicit IPs
    pre_existing_area = []
    for ddr in ddr_list:
      pre_existing_area.append(f'CLOCKREGION_X4Y{4 * ddr}:CLOCKREGION_X4Y{4 * ddr + 3}')

    # the vitis platform will take away the rightmost column
    if is_vitis_enabled:
      pre_existing_area.append(self.VITIS_REGION) # the area consumed by Vitis platform
    
    return pre_existing_area

  NAME = 'U280'
  FPGA_PART_NAME = 'xcu280-fsvh2892-2L-e'

  SLR_NUM = 4 # add a pseudo SLR at the top with area 0
  CR_NUM_HORIZONTAL = 8
  CR_NUM_VERTICAL = 12
  CR_NUM_VERTICAL_PER_SLR = 4
  
  TOTAL_AREA = {}
  TOTAL_AREA['BRAM'] = 4032
  TOTAL_AREA['DSP'] = 9024
  TOTAL_AREA['FF'] = 2607360
  TOTAL_AREA['LUT'] = 1303680
  TOTAL_AREA['URAM'] = 960

  # Clock Region level
  CR_AREA = defaultdict(lambda: defaultdict(dict))

  CR_AREA[0][0 ]['FF']  = 29760    
  CR_AREA[1][0 ]['FF']  = 24960 
  CR_AREA[0][1 ]['FF']  = 29760
  CR_AREA[1][1 ]['FF']  = 24960 
  CR_AREA[0][2 ]['FF']  = 29760
  CR_AREA[1][2 ]['FF']  = 24960 
  CR_AREA[0][3 ]['FF']  = 27840
  CR_AREA[1][3 ]['FF']  = 23040 
  CR_AREA[0][4 ]['FF']  = 27840
  CR_AREA[1][4 ]['FF']  = 23040 
  CR_AREA[0][5 ]['FF']  = 29760
  CR_AREA[1][5 ]['FF']  = 24960 
  CR_AREA[0][6 ]['FF']  = 29760
  CR_AREA[1][6 ]['FF']  = 24960 
  CR_AREA[0][7 ]['FF']  = 27840
  CR_AREA[1][7 ]['FF']  = 23040 
  CR_AREA[0][8 ]['FF']  = 27840
  CR_AREA[1][8 ]['FF']  = 23040 
  CR_AREA[0][9 ]['FF']  = 29760
  CR_AREA[1][9 ]['FF']  = 24960 
  CR_AREA[0][10]['FF']  =	29760
  CR_AREA[1][10]['FF']  =	24960 
  CR_AREA[0][11]['FF']  =	27840
  CR_AREA[1][11]['FF']  =	23040 
  CR_AREA[2][0 ]['FF']  = 36480
  CR_AREA[3][0 ]['FF']  = 21120 
  CR_AREA[2][1 ]['FF']  = 36480 
  CR_AREA[3][1 ]['FF']  = 21120 
  CR_AREA[2][2 ]['FF']  = 36480 
  CR_AREA[3][2 ]['FF']  = 21120 
  CR_AREA[2][3 ]['FF']  = 34560 
  CR_AREA[3][3 ]['FF']  = 19200
  CR_AREA[2][4 ]['FF']  = 34560 
  CR_AREA[3][4 ]['FF']  = 19200
  CR_AREA[2][5 ]['FF']  = 36480 
  CR_AREA[3][5 ]['FF']  = 21120 
  CR_AREA[2][6 ]['FF']  = 36480 
  CR_AREA[3][6 ]['FF']  = 21120 
  CR_AREA[2][7 ]['FF']  = 34560 
  CR_AREA[3][7 ]['FF']  = 19200
  CR_AREA[2][8 ]['FF']  = 34560 
  CR_AREA[3][8 ]['FF']  = 19200
  CR_AREA[2][9 ]['FF']  = 36480 
  CR_AREA[3][9 ]['FF']  = 21120 
  CR_AREA[2][10]['FF']  =	36480 
  CR_AREA[3][10]['FF']  =	21120 
  CR_AREA[2][11]['FF']  =	34560 
  CR_AREA[3][11]['FF']  =	19200
  CR_AREA[4][0 ]['FF']  = 27840 
  CR_AREA[5][0 ]['FF']  = 28800 
  CR_AREA[4][1 ]['FF']  = 27840 
  CR_AREA[5][1 ]['FF']  = 28800  
  CR_AREA[4][2 ]['FF']  = 27840 
  CR_AREA[5][2 ]['FF']  = 28800  
  CR_AREA[4][3 ]['FF']  = 25920 
  CR_AREA[5][3 ]['FF']  = 26880  
  CR_AREA[4][4 ]['FF']  = 25920 
  CR_AREA[5][4 ]['FF']  = 26880  
  CR_AREA[4][5 ]['FF']  = 27840 
  CR_AREA[5][5 ]['FF']  = 28800  
  CR_AREA[4][6 ]['FF']  = 27840 
  CR_AREA[5][6 ]['FF']  = 28800  
  CR_AREA[4][7 ]['FF']  = 25920 
  CR_AREA[5][7 ]['FF']  = 26880  
  CR_AREA[4][8 ]['FF']  = 25920 
  CR_AREA[5][8 ]['FF']  = 26880  
  CR_AREA[4][9 ]['FF']  = 27840 
  CR_AREA[5][9 ]['FF']  = 28800  
  CR_AREA[4][10]['FF']  =	27840 
  CR_AREA[5][10]['FF']  =	28800  
  CR_AREA[4][11]['FF']  =	25920 
  CR_AREA[5][11]['FF']  =	26880  
  CR_AREA[6][0 ]['FF']  = 28800 
  CR_AREA[7][0 ]['FF']  = 25920
  CR_AREA[6][1 ]['FF']  = 28800  
  CR_AREA[7][1 ]['FF']  = 25920
  CR_AREA[6][2 ]['FF']  = 28800  
  CR_AREA[7][2 ]['FF']  = 25920
  CR_AREA[6][3 ]['FF']  = 26880  
  CR_AREA[7][3 ]['FF']  = 24000
  CR_AREA[6][4 ]['FF']  = 26880  
  CR_AREA[7][4 ]['FF']  = 24000
  CR_AREA[6][5 ]['FF']  = 28800  
  CR_AREA[7][5 ]['FF']  = 25920
  CR_AREA[6][6 ]['FF']  = 28800  
  CR_AREA[7][6 ]['FF']  = 25920
  CR_AREA[6][7 ]['FF']  = 26880  
  CR_AREA[7][7 ]['FF']  = 24000
  CR_AREA[6][8 ]['FF']  = 26880  
  CR_AREA[7][8 ]['FF']  = 24000
  CR_AREA[6][9 ]['FF']  = 28800  
  CR_AREA[7][9 ]['FF']  = 25920
  CR_AREA[6][10]['FF']  =	28800  
  CR_AREA[7][10]['FF']  =	25920
  CR_AREA[6][11]['FF']  =	26880  
  CR_AREA[7][11]['FF']  =	24000
      
  ################    
      
  CR_AREA[0][0 ]['LUT'] = 14880  
  CR_AREA[1][0 ]['LUT'] = 12480   
  CR_AREA[0][1 ]['LUT'] = 14880  
  CR_AREA[1][1 ]['LUT'] = 12480   
  CR_AREA[0][2 ]['LUT'] = 14880  
  CR_AREA[1][2 ]['LUT'] = 12480   
  CR_AREA[0][3 ]['LUT'] = 13920  
  CR_AREA[1][3 ]['LUT'] = 11520   
  CR_AREA[0][4 ]['LUT'] = 13920  
  CR_AREA[1][4 ]['LUT'] = 11520   
  CR_AREA[0][5 ]['LUT'] = 14880  
  CR_AREA[1][5 ]['LUT'] = 12480   
  CR_AREA[0][6 ]['LUT'] = 14880  
  CR_AREA[1][6 ]['LUT'] = 12480   
  CR_AREA[0][7 ]['LUT'] = 13920  
  CR_AREA[1][7 ]['LUT'] = 11520   
  CR_AREA[0][8 ]['LUT'] = 13920  
  CR_AREA[1][8 ]['LUT'] = 11520   
  CR_AREA[0][9 ]['LUT'] = 14880  
  CR_AREA[1][9 ]['LUT'] = 12480   
  CR_AREA[0][10]['LUT'] = 14880  
  CR_AREA[1][10]['LUT'] = 12480   
  CR_AREA[0][11]['LUT'] = 13920  
  CR_AREA[1][11]['LUT'] = 11520   
  CR_AREA[2][0 ]['LUT'] = 18240  
  CR_AREA[3][0 ]['LUT'] = 10560   
  CR_AREA[2][1 ]['LUT'] = 18240   
  CR_AREA[3][1 ]['LUT'] = 10560   
  CR_AREA[2][2 ]['LUT'] = 18240   
  CR_AREA[3][2 ]['LUT'] = 10560   
  CR_AREA[2][3 ]['LUT'] = 17280   
  CR_AREA[3][3 ]['LUT'] = 9600	  
  CR_AREA[2][4 ]['LUT'] = 17280   
  CR_AREA[3][4 ]['LUT'] = 9600	  
  CR_AREA[2][5 ]['LUT'] = 18240   
  CR_AREA[3][5 ]['LUT'] = 10560   
  CR_AREA[2][6 ]['LUT'] = 18240   
  CR_AREA[3][6 ]['LUT'] = 10560   
  CR_AREA[2][7 ]['LUT'] = 17280   
  CR_AREA[3][7 ]['LUT'] = 9600	  
  CR_AREA[2][8 ]['LUT'] = 17280   
  CR_AREA[3][8 ]['LUT'] = 9600	  
  CR_AREA[2][9 ]['LUT'] = 18240   
  CR_AREA[3][9 ]['LUT'] = 10560   
  CR_AREA[2][10]['LUT'] = 18240   
  CR_AREA[3][10]['LUT'] = 10560   
  CR_AREA[2][11]['LUT'] = 17280   
  CR_AREA[3][11]['LUT'] = 9600	  
  CR_AREA[4][0 ]['LUT'] = 13920   
  CR_AREA[5][0 ]['LUT'] = 14400   
  CR_AREA[4][1 ]['LUT'] = 13920   
  CR_AREA[5][1 ]['LUT'] = 14400    
  CR_AREA[4][2 ]['LUT'] = 13920   
  CR_AREA[5][2 ]['LUT'] = 14400    
  CR_AREA[4][3 ]['LUT'] = 12960   
  CR_AREA[5][3 ]['LUT'] = 13440    
  CR_AREA[4][4 ]['LUT'] = 12960   
  CR_AREA[5][4 ]['LUT'] = 13440    
  CR_AREA[4][5 ]['LUT'] = 13920   
  CR_AREA[5][5 ]['LUT'] = 14400    
  CR_AREA[4][6 ]['LUT'] = 13920   
  CR_AREA[5][6 ]['LUT'] = 14400    
  CR_AREA[4][7 ]['LUT'] = 12960   
  CR_AREA[5][7 ]['LUT'] = 13440    
  CR_AREA[4][8 ]['LUT'] = 12960   
  CR_AREA[5][8 ]['LUT'] = 13440    
  CR_AREA[4][9 ]['LUT'] = 13920   
  CR_AREA[5][9 ]['LUT'] = 14400    
  CR_AREA[4][10]['LUT'] = 13920   
  CR_AREA[5][10]['LUT'] = 14400    
  CR_AREA[4][11]['LUT'] = 12960   
  CR_AREA[5][11]['LUT'] = 13440    
  CR_AREA[6][0 ]['LUT'] = 14400   
  CR_AREA[7][0 ]['LUT'] = 12960  
  CR_AREA[6][1 ]['LUT'] = 14400    
  CR_AREA[7][1 ]['LUT'] = 12960  
  CR_AREA[6][2 ]['LUT'] = 14400    
  CR_AREA[7][2 ]['LUT'] = 12960  
  CR_AREA[6][3 ]['LUT'] = 13440    
  CR_AREA[7][3 ]['LUT'] = 12000  
  CR_AREA[6][4 ]['LUT'] = 13440    
  CR_AREA[7][4 ]['LUT'] = 12000  
  CR_AREA[6][5 ]['LUT'] = 14400    
  CR_AREA[7][5 ]['LUT'] = 12960  
  CR_AREA[6][6 ]['LUT'] = 14400    
  CR_AREA[7][6 ]['LUT'] = 12960  
  CR_AREA[6][7 ]['LUT'] = 13440    
  CR_AREA[7][7 ]['LUT'] = 12000  
  CR_AREA[6][8 ]['LUT'] = 13440    
  CR_AREA[7][8 ]['LUT'] = 12000  
  CR_AREA[6][9 ]['LUT'] = 14400    
  CR_AREA[7][9 ]['LUT'] = 12960  
  CR_AREA[6][10]['LUT'] = 14400    
  CR_AREA[7][10]['LUT'] = 12960  
  CR_AREA[6][11]['LUT'] = 13440    
  CR_AREA[7][11]['LUT'] = 12000  
      
  ##################    
      
  CR_AREA[0][0 ]['DSP'] =	72	   
  CR_AREA[1][0 ]['DSP'] =	72	    
  CR_AREA[0][1 ]['DSP'] =	96	   
  CR_AREA[1][1 ]['DSP'] =	96	    
  CR_AREA[0][2 ]['DSP'] =	96	   
  CR_AREA[1][2 ]['DSP'] =	96	    
  CR_AREA[0][3 ]['DSP'] =	96	   
  CR_AREA[1][3 ]['DSP'] =	96	    
  CR_AREA[0][4 ]['DSP'] =	96	   
  CR_AREA[1][4 ]['DSP'] =	96	    
  CR_AREA[0][5 ]['DSP'] =	96	   
  CR_AREA[1][5 ]['DSP'] =	96	    
  CR_AREA[0][6 ]['DSP'] =	96	   
  CR_AREA[1][6 ]['DSP'] =	96	    
  CR_AREA[0][7 ]['DSP'] =	96	   
  CR_AREA[1][7 ]['DSP'] =	96	    
  CR_AREA[0][8 ]['DSP'] =	96	   
  CR_AREA[1][8 ]['DSP'] =	96	    
  CR_AREA[0][9 ]['DSP'] =	96	   
  CR_AREA[1][9 ]['DSP'] =	96	    
  CR_AREA[0][10]['DSP'] =	96	   
  CR_AREA[1][10]['DSP'] =	96	    
  CR_AREA[0][11]['DSP'] =	96	   
  CR_AREA[1][11]['DSP'] =	96	    
  CR_AREA[2][0 ]['DSP'] =	90	   
  CR_AREA[3][0 ]['DSP'] =	54	    
  CR_AREA[2][1 ]['DSP'] =	120	   
  CR_AREA[3][1 ]['DSP'] =	72	    
  CR_AREA[2][2 ]['DSP'] =	120	   
  CR_AREA[3][2 ]['DSP'] =	72	    
  CR_AREA[2][3 ]['DSP'] =	120	   
  CR_AREA[3][3 ]['DSP'] =	72	    
  CR_AREA[2][4 ]['DSP'] =	120	   
  CR_AREA[3][4 ]['DSP'] =	72	    
  CR_AREA[2][5 ]['DSP'] =	120	   
  CR_AREA[3][5 ]['DSP'] =	72	    
  CR_AREA[2][6 ]['DSP'] =	120	   
  CR_AREA[3][6 ]['DSP'] =	72	    
  CR_AREA[2][7 ]['DSP'] =	120	   
  CR_AREA[3][7 ]['DSP'] =	72	    
  CR_AREA[2][8 ]['DSP'] =	120	   
  CR_AREA[3][8 ]['DSP'] =	72	    
  CR_AREA[2][9 ]['DSP'] =	120	   
  CR_AREA[3][9 ]['DSP'] =	72	    
  CR_AREA[2][10]['DSP'] =	120	   
  CR_AREA[3][10]['DSP'] =	72	    
  CR_AREA[2][11]['DSP'] =	120	   
  CR_AREA[3][11]['DSP'] =	72	    
  CR_AREA[4][0 ]['DSP'] =	72	    
  CR_AREA[5][0 ]['DSP'] =	90	    
  CR_AREA[4][1 ]['DSP'] =	96	    
  CR_AREA[5][1 ]['DSP'] =	120	    
  CR_AREA[4][2 ]['DSP'] =	96	    
  CR_AREA[5][2 ]['DSP'] =	120	    
  CR_AREA[4][3 ]['DSP'] =	96	    
  CR_AREA[5][3 ]['DSP'] =	120	    
  CR_AREA[4][4 ]['DSP'] =	96	    
  CR_AREA[5][4 ]['DSP'] =	120	    
  CR_AREA[4][5 ]['DSP'] =	96	    
  CR_AREA[5][5 ]['DSP'] =	120	    
  CR_AREA[4][6 ]['DSP'] =	96	    
  CR_AREA[5][6 ]['DSP'] =	120	    
  CR_AREA[4][7 ]['DSP'] =	96	    
  CR_AREA[5][7 ]['DSP'] =	120	    
  CR_AREA[4][8 ]['DSP'] =	96	    
  CR_AREA[5][8 ]['DSP'] =	120	    
  CR_AREA[4][9 ]['DSP'] =	96	    
  CR_AREA[5][9 ]['DSP'] =	120	    
  CR_AREA[4][10]['DSP'] =	96	    
  CR_AREA[5][10]['DSP'] =	120	    
  CR_AREA[4][11]['DSP'] =	96	    
  CR_AREA[5][11]['DSP'] =	120	    
  CR_AREA[6][0 ]['DSP'] =	90	    
  CR_AREA[7][0 ]['DSP'] =	36	   
  CR_AREA[6][1 ]['DSP'] =	120	    
  CR_AREA[7][1 ]['DSP'] =	48	   
  CR_AREA[6][2 ]['DSP'] =	120	    
  CR_AREA[7][2 ]['DSP'] =	48	   
  CR_AREA[6][3 ]['DSP'] =	120	    
  CR_AREA[7][3 ]['DSP'] =	48	   
  CR_AREA[6][4 ]['DSP'] =	120	    
  CR_AREA[7][4 ]['DSP'] =	48	   
  CR_AREA[6][5 ]['DSP'] =	120	    
  CR_AREA[7][5 ]['DSP'] =	48	   
  CR_AREA[6][6 ]['DSP'] =	120	    
  CR_AREA[7][6 ]['DSP'] =	48	   
  CR_AREA[6][7 ]['DSP'] =	120	    
  CR_AREA[7][7 ]['DSP'] =	48	   
  CR_AREA[6][8 ]['DSP'] =	120	    
  CR_AREA[7][8 ]['DSP'] =	48	   
  CR_AREA[6][9 ]['DSP'] =	120	    
  CR_AREA[7][9 ]['DSP'] =	48	   
  CR_AREA[6][10]['DSP'] =	120	    
  CR_AREA[7][10]['DSP'] =	48	   
  CR_AREA[6][11]['DSP'] =	120	    
  CR_AREA[7][11]['DSP'] =	48	   
      
  ##################    
      
  CR_AREA[0][0 ]['BRAM'] = 24 * 2 
  CR_AREA[1][0 ]['BRAM'] = 24 * 2  
  CR_AREA[0][1 ]['BRAM'] = 24 * 2 
  CR_AREA[1][1 ]['BRAM'] = 24 * 2  
  CR_AREA[0][2 ]['BRAM'] = 24 * 2 
  CR_AREA[1][2 ]['BRAM'] = 24 * 2  
  CR_AREA[0][3 ]['BRAM'] = 24 * 2 
  CR_AREA[1][3 ]['BRAM'] = 24 * 2  
  CR_AREA[0][4 ]['BRAM'] = 24 * 2 
  CR_AREA[1][4 ]['BRAM'] = 24 * 2  
  CR_AREA[0][5 ]['BRAM'] = 24 * 2 
  CR_AREA[1][5 ]['BRAM'] = 24 * 2  
  CR_AREA[0][6 ]['BRAM'] = 24 * 2 
  CR_AREA[1][6 ]['BRAM'] = 24 * 2  
  CR_AREA[0][7 ]['BRAM'] = 24 * 2 
  CR_AREA[1][7 ]['BRAM'] = 24 * 2  
  CR_AREA[0][8 ]['BRAM'] = 24 * 2 
  CR_AREA[1][8 ]['BRAM'] = 24 * 2  
  CR_AREA[0][9 ]['BRAM'] = 24 * 2 
  CR_AREA[1][9 ]['BRAM'] = 24 * 2  
  CR_AREA[0][10]['BRAM'] = 24 * 2 
  CR_AREA[1][10]['BRAM'] = 24 * 2  
  CR_AREA[0][11]['BRAM'] = 24 * 2 
  CR_AREA[1][11]['BRAM'] = 24 * 2  
  CR_AREA[2][0 ]['BRAM'] = 36 * 2 
  CR_AREA[3][0 ]['BRAM'] = 12 * 2  
  CR_AREA[2][1 ]['BRAM'] = 36 * 2 
  CR_AREA[3][1 ]['BRAM'] = 12 * 2  
  CR_AREA[2][2 ]['BRAM'] = 36 * 2 
  CR_AREA[3][2 ]['BRAM'] = 12 * 2  
  CR_AREA[2][3 ]['BRAM'] = 36 * 2 
  CR_AREA[3][3 ]['BRAM'] = 12 * 2  
  CR_AREA[2][4 ]['BRAM'] = 36 * 2 
  CR_AREA[3][4 ]['BRAM'] = 12 * 2  
  CR_AREA[2][5 ]['BRAM'] = 36 * 2 
  CR_AREA[3][5 ]['BRAM'] = 12 * 2  
  CR_AREA[2][6 ]['BRAM'] = 36 * 2 
  CR_AREA[3][6 ]['BRAM'] = 12 * 2  
  CR_AREA[2][7 ]['BRAM'] = 36 * 2 
  CR_AREA[3][7 ]['BRAM'] = 12 * 2  
  CR_AREA[2][8 ]['BRAM'] = 36 * 2 
  CR_AREA[3][8 ]['BRAM'] = 12 * 2  
  CR_AREA[2][9 ]['BRAM'] = 36 * 2 
  CR_AREA[3][9 ]['BRAM'] = 12 * 2  
  CR_AREA[2][10]['BRAM'] = 36 * 2 
  CR_AREA[3][10]['BRAM'] = 12 * 2  
  CR_AREA[2][11]['BRAM'] = 36 * 2 
  CR_AREA[3][11]['BRAM'] = 12 * 2  
  CR_AREA[4][0 ]['BRAM'] = 24 * 2  
  CR_AREA[5][0 ]['BRAM'] = 12 * 2  
  CR_AREA[4][1 ]['BRAM'] = 24 * 2  
  CR_AREA[5][1 ]['BRAM'] = 12 * 2  
  CR_AREA[4][2 ]['BRAM'] = 24 * 2  
  CR_AREA[5][2 ]['BRAM'] = 12 * 2  
  CR_AREA[4][3 ]['BRAM'] = 24 * 2  
  CR_AREA[5][3 ]['BRAM'] = 12 * 2  
  CR_AREA[4][4 ]['BRAM'] = 24 * 2  
  CR_AREA[5][4 ]['BRAM'] = 12 * 2  
  CR_AREA[4][5 ]['BRAM'] = 24 * 2  
  CR_AREA[5][5 ]['BRAM'] = 12 * 2  
  CR_AREA[4][6 ]['BRAM'] = 24 * 2  
  CR_AREA[5][6 ]['BRAM'] = 12 * 2  
  CR_AREA[4][7 ]['BRAM'] = 24 * 2  
  CR_AREA[5][7 ]['BRAM'] = 12 * 2  
  CR_AREA[4][8 ]['BRAM'] = 24 * 2  
  CR_AREA[5][8 ]['BRAM'] = 12 * 2  
  CR_AREA[4][9 ]['BRAM'] = 24 * 2  
  CR_AREA[5][9 ]['BRAM'] = 12 * 2  
  CR_AREA[4][10]['BRAM'] = 24 * 2  
  CR_AREA[5][10]['BRAM'] = 12 * 2  
  CR_AREA[4][11]['BRAM'] = 24 * 2  
  CR_AREA[5][11]['BRAM'] = 12 * 2  
  CR_AREA[6][0 ]['BRAM'] = 12 * 2  
  CR_AREA[7][0 ]['BRAM'] = 24 * 2 
  CR_AREA[6][1 ]['BRAM'] = 12 * 2  
  CR_AREA[7][1 ]['BRAM'] = 24 * 2 
  CR_AREA[6][2 ]['BRAM'] = 12 * 2  
  CR_AREA[7][2 ]['BRAM'] = 24 * 2 
  CR_AREA[6][3 ]['BRAM'] = 12 * 2  
  CR_AREA[7][3 ]['BRAM'] = 24 * 2 
  CR_AREA[6][4 ]['BRAM'] = 12 * 2  
  CR_AREA[7][4 ]['BRAM'] = 24 * 2 
  CR_AREA[6][5 ]['BRAM'] = 12 * 2  
  CR_AREA[7][5 ]['BRAM'] = 24 * 2 
  CR_AREA[6][6 ]['BRAM'] = 12 * 2  
  CR_AREA[7][6 ]['BRAM'] = 24 * 2 
  CR_AREA[6][7 ]['BRAM'] = 12 * 2  
  CR_AREA[7][7 ]['BRAM'] = 24 * 2 
  CR_AREA[6][8 ]['BRAM'] = 12 * 2  
  CR_AREA[7][8 ]['BRAM'] = 24 * 2 
  CR_AREA[6][9 ]['BRAM'] = 12 * 2  
  CR_AREA[7][9 ]['BRAM'] = 24 * 2 
  CR_AREA[6][10]['BRAM'] = 12 * 2  
  CR_AREA[7][10]['BRAM'] = 24 * 2 
  CR_AREA[6][11]['BRAM'] = 12 * 2  
  CR_AREA[7][11]['BRAM'] = 24 * 2 
      
  #####################    
      
  CR_AREA[0][0 ]['URAM'] = 0    
  CR_AREA[1][0 ]['URAM'] = 16    
  CR_AREA[0][1 ]['URAM'] = 0    
  CR_AREA[1][1 ]['URAM'] = 16    
  CR_AREA[0][2 ]['URAM'] = 0    
  CR_AREA[1][2 ]['URAM'] = 16    
  CR_AREA[0][3 ]['URAM'] = 0    
  CR_AREA[1][3 ]['URAM'] = 16    
  CR_AREA[0][4 ]['URAM'] = 0    
  CR_AREA[1][4 ]['URAM'] = 16    
  CR_AREA[0][5 ]['URAM'] = 0    
  CR_AREA[1][5 ]['URAM'] = 16    
  CR_AREA[0][6 ]['URAM'] = 0    
  CR_AREA[1][6 ]['URAM'] = 16    
  CR_AREA[0][7 ]['URAM'] = 0    
  CR_AREA[1][7 ]['URAM'] = 16    
  CR_AREA[0][8 ]['URAM'] = 0    
  CR_AREA[1][8 ]['URAM'] = 16    
  CR_AREA[0][9 ]['URAM'] = 0    
  CR_AREA[1][9 ]['URAM'] = 16    
  CR_AREA[0][10]['URAM'] = 0    
  CR_AREA[1][10]['URAM'] = 16    
  CR_AREA[0][11]['URAM'] = 0    
  CR_AREA[1][11]['URAM'] = 16    
  CR_AREA[2][0 ]['URAM'] = 0    
  CR_AREA[3][0 ]['URAM'] = 16    
  CR_AREA[2][1 ]['URAM'] = 0    
  CR_AREA[3][1 ]['URAM'] = 16    
  CR_AREA[2][2 ]['URAM'] = 0    
  CR_AREA[3][2 ]['URAM'] = 16    
  CR_AREA[2][3 ]['URAM'] = 0    
  CR_AREA[3][3 ]['URAM'] = 16    
  CR_AREA[2][4 ]['URAM'] = 0    
  CR_AREA[3][4 ]['URAM'] = 16    
  CR_AREA[2][5 ]['URAM'] = 0    
  CR_AREA[3][5 ]['URAM'] = 16    
  CR_AREA[2][6 ]['URAM'] = 0    
  CR_AREA[3][6 ]['URAM'] = 16    
  CR_AREA[2][7 ]['URAM'] = 0    
  CR_AREA[3][7 ]['URAM'] = 16    
  CR_AREA[2][8 ]['URAM'] = 0    
  CR_AREA[3][8 ]['URAM'] = 16    
  CR_AREA[2][9 ]['URAM'] = 0    
  CR_AREA[3][9 ]['URAM'] = 16    
  CR_AREA[2][10]['URAM'] = 0    
  CR_AREA[3][10]['URAM'] = 16    
  CR_AREA[2][11]['URAM'] = 0    
  CR_AREA[3][11]['URAM'] = 16    
  CR_AREA[4][0 ]['URAM'] = 16    
  CR_AREA[5][0 ]['URAM'] = 16    
  CR_AREA[4][1 ]['URAM'] = 16    
  CR_AREA[5][1 ]['URAM'] = 16    
  CR_AREA[4][2 ]['URAM'] = 16    
  CR_AREA[5][2 ]['URAM'] = 16    
  CR_AREA[4][3 ]['URAM'] = 16    
  CR_AREA[5][3 ]['URAM'] = 16    
  CR_AREA[4][4 ]['URAM'] = 16    
  CR_AREA[5][4 ]['URAM'] = 16    
  CR_AREA[4][5 ]['URAM'] = 16    
  CR_AREA[5][5 ]['URAM'] = 16    
  CR_AREA[4][6 ]['URAM'] = 16    
  CR_AREA[5][6 ]['URAM'] = 16    
  CR_AREA[4][7 ]['URAM'] = 16    
  CR_AREA[5][7 ]['URAM'] = 16    
  CR_AREA[4][8 ]['URAM'] = 16    
  CR_AREA[5][8 ]['URAM'] = 16    
  CR_AREA[4][9 ]['URAM'] = 16    
  CR_AREA[5][9 ]['URAM'] = 16    
  CR_AREA[4][10]['URAM'] = 16    
  CR_AREA[5][10]['URAM'] = 16    
  CR_AREA[4][11]['URAM'] = 16    
  CR_AREA[5][11]['URAM'] = 16    
  CR_AREA[6][0 ]['URAM'] = 16    
  CR_AREA[7][0 ]['URAM'] = 0    
  CR_AREA[6][1 ]['URAM'] = 16    
  CR_AREA[7][1 ]['URAM'] = 0    
  CR_AREA[6][2 ]['URAM'] = 16    
  CR_AREA[7][2 ]['URAM'] = 0    
  CR_AREA[6][3 ]['URAM'] = 16    
  CR_AREA[7][3 ]['URAM'] = 0    
  CR_AREA[6][4 ]['URAM'] = 16    
  CR_AREA[7][4 ]['URAM'] = 0    
  CR_AREA[6][5 ]['URAM'] = 16    
  CR_AREA[7][5 ]['URAM'] = 0    
  CR_AREA[6][6 ]['URAM'] = 16    
  CR_AREA[7][6 ]['URAM'] = 0    
  CR_AREA[6][7 ]['URAM'] = 16    
  CR_AREA[7][7 ]['URAM'] = 0    
  CR_AREA[6][8 ]['URAM'] = 16    
  CR_AREA[7][8 ]['URAM'] = 0    
  CR_AREA[6][9 ]['URAM'] = 16    
  CR_AREA[7][9 ]['URAM'] = 0    
  CR_AREA[6][10]['URAM'] = 16    
  CR_AREA[7][10]['URAM'] = 0    
  CR_AREA[6][11]['URAM'] = 16    
  CR_AREA[7][11]['URAM'] = 0

  # for item in ('BRAM', 'DSP', 'FF', 'LUT', 'URAM'):
  #   assert TOTAL_AREA[item] == sum(CR_AREA[x][y][item] for x in range(8) for y in range(12))

  VITIS_REGION = f'CLOCKREGION_X7Y1:CLOCKREGION_X7Y11'

  DDR_TO_CLOCK_REGIONS = {
    0: 'CLOCKREGION_X4Y1:CLOCKREGION_X4Y3',
    1: 'CLOCKREGION_X4Y5:CLOCKREGION_X4Y7',
  }

  def getDDRSlolenRegion(self, ddr: int) -> str:
    assert 0 <= ddr <= 3, ddr
    # entire sloten region is {SLICE_X144Y0:SLICE_X145Y959 DSP48E2_X19Y0:DSP48E2_X19Y383 RAMB18_X9Y0:RAMB18_X9Y383 RAMB36_X9Y0:RAMB36_X9Y191}
    return f'SLICE_X144Y{240 * ddr}:SLICE_X145Y{240 * (ddr+1) - 1} DSP48E2_X19Y{96 * ddr}:DSP48E2_X19Y{96 * (ddr+1) - 1} RAMB18_X9Y{96 * ddr}:RAMB18_X9Y{96 * (ddr+1) - 1} RAMB36_X9Y{48 * ddr}:RAMB36_X9Y{48 * (ddr+1) - 1}'


class DeviceManager:
  def __init__(self, board_name, ddr_list = [], is_vitis_enabled = False):
    if board_name == 'U250':
      self.board = DeviceU250(ddr_list, is_vitis_enabled)
    elif board_name == 'U280':
      self.board = DeviceU280(ddr_list, is_vitis_enabled)
    else:
      assert False, f'unsupported device: {board_name}'

  def getBoard(self):
    return self.board